#include <math.h>
#include <string.h>
#include "encode.h"

#define NULL 0

#define PADDING_BYTE 2
#define PADDING_BIT	0x02
#define STEREO_BYTE 3
#define STEREO_MASK 0xf0

#define JOINT_STERIO_ELECTION_LENGTH 50
#define JOINT_STERIO_VOTE_THRESHOLD 25

CAudioEncode::CAudioEncode(
				ESamplingFreq Freq /*=SF_44100_HZ*/,
				EStereoType Type /*= STEREO*/,
				int ReqBitRate /*= 160 (in KHz)*/)  : RightDelay(INPUTSIZE,256-64),
													  LeftDelay(INPUTSIZE,256-64),
                                                      CalcSMR(Freq, ReqBitRate)
/*
audio encode constructor.  Store compression parameters
and calculate output number of bytes.
*/
{
	int i;
    int BitRate;

	// Force a first time vote
    JointSterioVoteCount = -1;
	JointSterioVotes = 0;


	SamplingFreqCode = Freq;
    if (SamplingFreqCode > SF_11025_HZ)
         SamplingFreqCode = SF_11025_HZ;
	StereoType = Type;
    if (StereoType > MONO)
        StereoType = STEREO;

	//compute output size variables
	switch (SamplingFreqCode) {
	case SF_48000_HZ:
		SamplingFreq = 48000;
		break;
	case SF_32000_HZ:
		SamplingFreq = 32000;
		break;
	case SF_44100_HZ:
	case SF_22050_HZ:
	case SF_11025_HZ:
		SamplingFreqCode = SF_44100_HZ;
		SamplingFreq = 44100;
		break;
	default:
		SamplingFreq = 44100;
	}

	//clip requested bit rate to max/min allowed
	int MaxBitRateCode = sizeof(BitRates)/sizeof(BitRates[0]) - 1;
	BitRate = min(ReqBitRate,BitRates[MaxBitRateCode].rate);
	BitRate = max(ReqBitRate,BitRates[1].rate);

	//find bit rate code in table or use free rate (index 0) 
	int modebit = 1 << (int)StereoType;
	for (i=MaxBitRateCode; i>0; i--) {
		if (BitRates[i].rate == BitRate && (BitRates[i].ModeUsed & modebit))
			break;
	}
	BitRateCode = i;

    //init pointer to current allocation info table
    int PerChanRate;
    if (StereoType == MONO)
        PerChanRate = BitRate;
    else
        PerChanRate = BitRate/2;
    int TableSize = sizeof(AllocationSetup)/sizeof(AllocationSetup[0]);
    for (i = TableSize-1; i>0; i--) {    //look up pre channel bit rate in allocaton setup table
        if (PerChanRate == AllocationSetup[i].rate)
            break;
    }
    

	//look up allocation info pointer in allocaton setup table based on sampling freq
    AllocInfo = AllocationSetup[i].AllocInfo[(int)SamplingFreqCode];

	FrameSize = (INPUTSIZE/8) * BitRate * 1000 / SamplingFreq;   //Output frame size (in bytes)
	rest = 0;       					  //init frame size remainder and running remainder sum
	dif = (INPUTSIZE/8) * BitRate * 1000 % SamplingFreq; //(used to determine when extra byte is needed, see 11172-3 2.4.2.3)


	//init frame header
	header[0] = (BYTE)0xff;  //8 bits of sync
	header[1] = (BYTE)0xfd;   //4 more bits of sync, layer II, and no CRC
	header[2] = ((BYTE)BitRateCode << 4) + (((BYTE)SamplingFreqCode) << 2); //bit rate, sampling freq and padding bit
	header[3] = 0x00 + (StereoType << 6);  //set for no emphasis, copyright off and not original
									   //init mode & submode, may change if mode is joint stereo

	//init inverted scale factor table
	for (i=0; i<SCALEFACTORTABLESIZE; i++)
		InvertedScaleFactors[i] = 1.0F/ScaleFactorTable[i];

    //find sub band limit = number of used subbands
	for (i=0;i<SUBBANDS; i++) {
		if (AllocInfo->sequence[i]->NumBits == 0)
            break;
    }
    sbLimit = i;
	
	// Set Decimation and Subband upper limit based on the lower of input bw and bit allocations
	Decimation = 1;
	if (Freq == SF_22050_HZ) {
		Decimation = 2;
		if (i < 16)
			sbLimitVar = i;
		else
			sbLimitVar = 16; 
		sbLimitBW = sbLimitVar;
	}
	else if (Freq == SF_11025_HZ) {
		Decimation = 4;
		if (i < 8)
			sbLimitVar = i;
		else
			sbLimitVar = 8; 
		sbLimitBW = sbLimitVar;
	}
	else {
		sbLimitVar = i; 
		sbLimitBW = i;
	}

}

int CAudioEncode::GetNextFrameSize(void)
//returns the size of the next output frame in bytes
{
	if (rest < 0)
		return FrameSize + 1;
	else
		return FrameSize;
}

int CAudioEncode::GetMaxFrameSize(void)
//returns the size of the largest possible output frame
{
	if (dif)
		return FrameSize + 1;
	else
		return FrameSize;
}

void CAudioEncode::ProcessFrame(short *FrameInput, BYTE *FrameOutput,
								int combined_smr,int *MaxSB)
/*
Process the next frame of input data 

short *FrameInput,    1152 16 bit samples
					  or 2*1152 16 samples for stereo
					  (right0,left0, right1,left1 ...)
                      (1152/2 samples if sample rate == 22050)
                      (1152/4 samples if sample rate == 11025)

BYTE  *FrameOutput,   (1152 * (bit rate / 8) / sample rate) bytes on 
					  average (output size may vary by 1 byte)
int   combined_smr,   Compute SPR based on sum of both channels (faster)
int   *MaxSB);        MAX Subband to process (0= use default)
											  // default returned
*/
{
	int i,v;

	//convert input to floating point
	short *inptr = FrameInput;

	// Setup the amount of processing we are willing to perform
	if ((*MaxSB < 8) || (*MaxSB > sbLimitBW)) {
		*MaxSB = sbLimitBW;
	}
	sbLimitVar = *MaxSB;

	switch (Decimation) {

		// normal mode (every sample)
	case 1:
			if (StereoType == MONO) {
				//if mono data, only use right channel
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i++)
						RightData[v][i] = (float)*inptr++;
				}
			}
			else {
				//if stereo data, de-mux into right and left channels
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i++) {
						RightData[v][i] = (float)*inptr++;
						LeftData[v][i]  = (float)*inptr++;
					}
				}
			}
			break;

		// 22K up-sample by repeating every print twice
		case 2:
			if (StereoType == MONO) {
				//if mono data, only use right channel
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i += 2) {
						RightData[v][i] = (float)*inptr;
						RightData[v][i+1] = (float)*inptr++;
					}
				}
			}
			else {
				//if stereo data, de-mux into right and left channels
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i += 2) {
						RightData[v][i] = (float)*inptr;
						RightData[v][i+1] = (float)*inptr++;

						LeftData[v][i]  = (float)*inptr;
						LeftData[v][i+1]  = (float)*inptr++;
					}
				}
			}
			break;


		// 11K up-sample by repeating every point 4 times
		case 4:
			if (StereoType == MONO) {
				//if mono data, only use right channel
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i += 4) {
						RightData[v][i]   = (float)*inptr;
						RightData[v][i+1] = (float)*inptr;
						RightData[v][i+2] = (float)*inptr;
						RightData[v][i+3] = (float)*inptr++;
					}
				}
			}
			else {
				//if stereo data, de-mux into right and left channels
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i += 4) {
						RightData[v][i]   = (float)*inptr;
						RightData[v][i+1] = (float)*inptr;
						RightData[v][i+2] = (float)*inptr;
						RightData[v][i+3] = (float)*inptr++;

						LeftData[v][i]    = (float)*inptr;
						LeftData[v][i+1]  = (float)*inptr;
						LeftData[v][i+2]  = (float)*inptr;
						LeftData[v][i+3]  = (float)*inptr++;
					}
				}
			}
			break;

	} // switch (SamplingFreqCode) 

	//time to encode...
	ProcessCurData(combined_smr,FrameOutput);
}


void CAudioEncode::ProcessFrame(float (*FrameInput)[2], BYTE *FrameOutput,
								int combined_smr,int *MaxSB)
/*
Process the next frame of input data 

float (*FrameInput)[2]     stereo floating point input samples

BYTE  *FrameOutput) 	   (1152 * (bit rate / 8) / sample rate) bytes on 
					       average (output size may vary by 1 byte)
*/
{
	int i,v;

	float (*inptr)[2] = FrameInput;

	// Setup the amount of processing we are willing to perform
	if ((*MaxSB < 8) || (*MaxSB > sbLimitBW)) {
		*MaxSB = sbLimitBW;
	}
	sbLimitVar = *MaxSB;

	switch (Decimation) {

		// normal mode (every sample)
		
		case 1:
			if (StereoType == MONO) {
				//mono data, average input into right channel
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i++, inptr++)
						RightData[v][i] =  0.5F * ((*inptr)[0] + (*inptr)[1]);
				}
			}
			else {
				//stereo data: copy it in
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i++) {
						RightData[v][i] = (*inptr)[0];
						LeftData[v][i] = (*inptr++)[1];
					}
				}
			}
			break;

		case 2:
			if (StereoType == MONO) {
				//mono data, average input into right channel
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i+=2, inptr++) {
						RightData[v][i] =  0.5F * ((*inptr)[0] + (*inptr)[1]);
						RightData[v][i+1] =  RightData[v][i];
					}
				}
			}
			else {
				//stereo data: copy it in
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i+=2, inptr++) {
						RightData[v][i+1] = RightData[v][i] = (*inptr)[0];
						LeftData[v][i+1] = LeftData[v][i] = (*inptr)[1];
					}
				}
			}
			break;

		case 4:
			if (StereoType == MONO) {
				//mono data, average input into right channel
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i+=4, inptr++) {
						RightData[v][i] =  0.5F * ((*inptr)[0] + (*inptr)[1]);
						RightData[v][i+1] =  RightData[v][i];
						RightData[v][i+2] =  RightData[v][i];
						RightData[v][i+3] =  RightData[v][i];
					}
				}
			}
			else {
				//stereo data: copy it in
				for (v=0; v<VECTORS; v++) {
					for (i=0; i<SUBBANDS; i+=4, inptr++) {
						RightData[v][i] = (*inptr)[0];
						RightData[v][i+1] =  (*inptr)[0];
						RightData[v][i+2] =  (*inptr)[0];
						RightData[v][i+3] =  (*inptr)[0];
						LeftData[v][i] = (*inptr)[1];
						LeftData[v][i+1] =  (*inptr)[1];
						LeftData[v][i+2] =  (*inptr)[1];
						LeftData[v][i+3] =  (*inptr)[1];

					}
				}
			}
			break;

	} // switch (SamplingFreqCode)


	//do the ol' encode thing
	ProcessCurData(combined_smr,FrameOutput);
}


void CAudioEncode::ProcessCurData(int combined_smr,BYTE *FrameOutput)
/*
Process the next frame of input data which has is in
the RightData & LeftData member arrays.

Output:

BYTE  *FrameOutput);  (1152 * (bit rate / 8) / sample rate) bytes on 
					  average (output size may vary by 1 byte)
*/
{
	int i,v;

    if (combined_smr && (StereoType != MONO)) {

		float SumData[VECTORS][SUBBANDS];
		int SumScaleFactors[SUBBANDS][NUMGROUPS];
		int SumSelInfo[SUBBANDS];

		//delay raw data to match subband filter for SMR calculation
		RightDelay.AddData((float *)RightData);

		//use subband filters to compute frequency data
		for (i=0; i<VECTORS; i++) {
			RightFilter.CalcSubbands(sbLimitVar,RightData[i],RightData[i]);
		}

		//compute scalefactors
		CalcScaleFactors(RightData,RightScaleFactors,RightSelInfo);

		//delay raw data to match subband filter for SMR calculation
		LeftDelay.AddData((float *)LeftData);

		//use subband filters to compute frequency data
		for (i=0; i<VECTORS; i++) {
			LeftFilter.CalcSubbands(sbLimitVar,LeftData[i],LeftData[i]);
		}

		//compute scalefactors
		CalcScaleFactors(LeftData,LeftScaleFactors,LeftSelInfo);

		//Compute the scale factors based on the max of the two channels
		for (v=0; v<VECTORS; v++) 
			for (i=0; i<SUBBANDS; i++) 
				SumData[v][i] = 0.5F * (RightData[v][i] + LeftData[v][i]);
		

		// compute scale factors and save in both channels
		CalcScaleFactors(SumData,SumScaleFactors,SumSelInfo);

		// Compute the combined channel
		float *rp = RightDelay.GetData();
		float *lp = LeftDelay.GetData();
		float DelaySum[INPUTSIZE];
		for (i=0;i<INPUTSIZE;i++) {
			DelaySum[i] = 0.5F * (rp[i] + lp[i]);
		}

		//compute signal to mask ratio with psychoacoustic model
		CalcSMR.CalcSignalToMask(DelaySum, SumScaleFactors, RightSMR, sbLimitVar);
		memcpy(LeftSMR,RightSMR,sizeof(LeftSMR));
		
	}
	else {

		//delay raw data to match subband filter for SMR calculation
		RightDelay.AddData((float *)RightData);

		//use subband filters to compute frequency data
		for (i=0; i<VECTORS; i++) {
			RightFilter.CalcSubbands(sbLimitVar,RightData[i],RightData[i]);
		}

		//compute scalefactors
		CalcScaleFactors(RightData,RightScaleFactors,RightSelInfo);

		//compute signal to mask ratio with psychoacoustic model
		CalcSMR.CalcSignalToMask(RightDelay.GetData(), RightScaleFactors, RightSMR, sbLimitVar);
    
		//if stereo, also compute left channel
		if (StereoType != MONO) {
			//delay raw data to match subband filter for SMR calculation
			LeftDelay.AddData((float *)LeftData);

			//use subband filters to compute frequency data
			for (i=0; i<VECTORS; i++) {
				LeftFilter.CalcSubbands(sbLimitVar,LeftData[i],LeftData[i]);
			}
			//compute scalefactors
			CalcScaleFactors(LeftData,LeftScaleFactors,LeftSelInfo);

			//compute signal to mask ratio with psychoacoustic model
			CalcSMR.CalcSignalToMask(LeftDelay.GetData(), LeftScaleFactors, LeftSMR, sbLimitVar);
		}
	}

	//First time; set it on the first block
	if (JointSterioVoteCount == -1) {
		JointStereoBound = CalcJointStereoBound();
		JointSterioVoteCount = 0;
		JointSterioVotes = 0;
	}
	else {

		// Cast a vote on the new joint sterio value
		int new_bound = CalcJointStereoBound();
		if (new_bound > JointStereoBound)
			JointSterioVotes++;
		else if (new_bound < JointStereoBound)
			JointSterioVotes--;

		// Election over; make the adjustment
		JointSterioVoteCount++;
		if (JointSterioVoteCount >= JOINT_STERIO_ELECTION_LENGTH) {

			// Survey says; raise it
			if (JointSterioVotes >= JOINT_STERIO_VOTE_THRESHOLD) {
				JointStereoBound = IncrementBound(JointStereoBound);
			}
			   
			// No lower it
			else if (JointSterioVotes <= -(JOINT_STERIO_VOTE_THRESHOLD)) {
				JointStereoBound = DecrementBound(JointStereoBound);
			}

			// Start a new term
			JointSterioVoteCount = 0;
			JointSterioVotes = 0;

		}
	}

	
    //update signal to mask ratio for joint sterio
    if (StereoType == JOINT_STEREO) {
        for (i=JointStereoBound; i<sbLimit; i++)
            RightSMR[i] = max(RightSMR[i], LeftSMR[i]);
    }

    //how many bits does each subband get?  Let's find out
    CalcBitAllocation();

    //pack those bits into the output frame
    BuildOuptut(FrameOutput);

    //update frame size variables for next frame (see 11172-3 2.4.2.3)
    if (rest < 0)
        rest += SamplingFreq;
    rest -= dif;
	
}

int CAudioEncode::CalcScaleFactorClass(int sf1, int sf2)
/*
this helper function is used to compute the class of the specified scale factors,
see sec C.1.5.2.5.  Note that this implementation starts scale factor class numbering from
0, not 1 as in the spec
*/
{
	int sfdiff = sf1-sf2;
	if (sfdiff < 0) {
		if (sfdiff < -2)
			return 0;
		else
			return 1;
	}
	else if (sfdiff > 0) {
		if (sfdiff > 2)
			return 4;
		else
			return 3;
	}
	else	//sfdiff == 0
		return 2;
}


void CAudioEncode::CalcScaleFactors(float SubBandData[VECTORS][SUBBANDS],
									int ScaleFactors[SUBBANDS][NUMGROUPS],
									int SelInfo[SUBBANDS])
/*
This function computes the index in the scale factors table for the specifed subband data.
The input is in the form of 36 vectors of 32 subbands:
float SubBandData[GROUPSIZE * NUMGROUPS][SUBBANDS]

the output is in the form of 32 tri-values
int ScaleFactors[SUBBANDS][NUMGROUPS]

this routine also updates computes the scale factor selection info and updates the scale factors 
based on table C.4

*/
{
	int sb,gs,ng;

	//for each subband compute 3 scale factors
	for (sb=0; sb<sbLimit; sb++) {
		//for each vector group of 12 vectors

		int vec = 0;	 //init to 1st vector
		for (ng=0; ng<NUMGROUPS; ng++) {
			float maxabs = 0.0F;
			//find max abs value for each subband in this group of 12 vectors
			for (gs=0; gs<GROUPSIZE; gs++) {
				maxabs = max(maxabs,absval(SubBandData[vec++][sb]));
			}
			//find index of smallest table entry that is larger then maxabs
			//(the table entries decrease in value)
			int i;
			for (i=1; i<SCALEFACTORTABLESIZE; i++) {
				if (ScaleFactorTable[i] < maxabs)
					break;
			}
			//output scale factor index
			ScaleFactors[sb][ng] = i - 1;
		}

		//use group of 3 scalefactors for this subband to compute the  
		//scale factor selection info
		int class1 = CalcScaleFactorClass(ScaleFactors[sb][0], ScaleFactors[sb][1]);
		int class2 = CalcScaleFactorClass(ScaleFactors[sb][1], ScaleFactors[sb][2]);
		const SScaleFactorInfo *sfInfo = &ScaleFactorInfo[class1][class2];
		SelInfo[sb] = sfInfo->SelInfo;

		//update the scale factor data from table C.4
		if (sfInfo->sf0 < 0) {
		    //entry -1 means use max of 3 scalefactors (max SF at min index)
			int sfmax = min(ScaleFactors[sb][0],min(ScaleFactors[sb][1],ScaleFactors[sb][2]));
			ScaleFactors[sb][0] = sfmax;
			ScaleFactors[sb][1] = sfmax;
			ScaleFactors[sb][2] = sfmax;
		}
		else {
			// otherwise, replace based on the table
			ScaleFactors[sb][0] = ScaleFactors[sb][sfInfo->sf0];
			ScaleFactors[sb][1] = ScaleFactors[sb][sfInfo->sf1];
			ScaleFactors[sb][2] = ScaleFactors[sb][sfInfo->sf2];
		}

	} //end for each subband

}



int CAudioEncode::CalcJointStereoBound(void)
/*
This function uses the object's signal to mask, and output frame size data
to compute the best choice for the bound in joint stereo mode, one of:
	bound=4  ==> subbands  4-31 in intensity stereo
	bound=8  ==> subbands  8-31 in intensity stereo
	bound=12 ==> subbands 12-31 in intensity stereo
	bound=16 ==> subbands 16-31 in intensity stereo
	bound=32 ==> use stereo mode for this frame

This routine retunrs the highest bound that results in "no signal degradation"
(eg when the Signal to noise - Signal to mask >= 0) for the available output bits.
If no bound yields "no signal degradation", bound=4 is returned.

If mono mode is set, the routine returns 0.  If stereo of dual channel mod is set the
routine returns sbLimit.
*/
{
    if (StereoType == MONO)
        return 0;
    if (StereoType != JOINT_STEREO)
        return sbLimit;

	//init available bits
	int DataBits = GetNextFrameSize() * 8;

    //1st, compute the data bits that are required for each subband at "no degradation"
    int RightBits[SUBBANDS];
    int LeftBits[SUBBANDS];
    int i;
    for (i=0; i<sbLimit; i++) {
        RightBits[i] = SMRToBits(i,RightSMR[i]);
        LeftBits[i] = SMRToBits(i,LeftSMR[i]);
    }
    for (; i<SUBBANDS; i++) 
        RightBits[i] = LeftBits[i] = 0;
	
	int bound = sbLimit;
	if (GetRequiredBits(bound, RightBits, LeftBits) <= DataBits)
		return bound;
	
	bound = min (16, sbLimit); 
	if (GetRequiredBits(bound, RightBits, LeftBits) <= DataBits)
		return bound;

	bound = min (12, sbLimit); 
	if (GetRequiredBits(bound, RightBits, LeftBits) <= DataBits)
		return bound;

	bound = min (8, sbLimit); 
	if (GetRequiredBits(bound, RightBits, LeftBits) <= DataBits)
		return bound;

	bound = min (4, sbLimit); 
	return bound;

}

// Select the next higher JountSterio sublimit bound
int CAudioEncode::IncrementBound (int current_bound) {
	switch (current_bound) {
		case 4:
			return(8);
		case 8:
			return(12);
		case 12:
			return(16);
		case 16:
			return(sbLimit);
		default:
			return(sbLimit);
	}
}

// Select the next Lower JountSterio sublimit bound
int CAudioEncode::DecrementBound (int current_bound) {
	switch (current_bound) {
		case 4:
			return(4);
		case 8:
			return(4);
		case 12:
			return(8);
		case 16:
			return(12);
		default:
			return(16);
	}
}


int CAudioEncode::GetRequiredBits(int bound, int RightBits[SUBBANDS], int LeftBits[SUBBANDS])
/*
This function returns the number of required bits to encode the current frame of data
at the specified joint stereo bound.  The 2 array inputs are:
  int RightBits[SUBBANDS]       number of bits to encode sample data for each right subband  
  int LeftBits[SUBBANDS]        number of bits to encode sample data for each left subband 
  
this funciton also uses the scale factor selection info (RightSelInfo & LeftSelInfo)
to compute the number of bits needed

*/
{
	int i;

	//compute header bits
	int ReqBits = sizeof(header) * 8;

	//compute allocation bits
	for(i=0; i<bound; i++)
		ReqBits += 2*AllocInfo->sequence[i]->NumBits;
	for(; i<sbLimit; i++)
		ReqBits += AllocInfo->sequence[i]->NumBits;

    //compute data bits
	for (i=0; i<bound; i++)
        ReqBits += RightBits[i] + LeftBits[i];

    for (; i<sbLimit; i++)
        ReqBits += max(RightBits[i], LeftBits[i]);

    //compute scale factor and scale factor selectoin bits
    for (i=0; i<bound; i++) {
        if (RightBits[i])  //if we send data, then we need scale factors
            ReqBits +=  SelInfoToBits(RightSelInfo[i]) + 2; // +2 for selection bits
        if (LeftBits[i])   //if we send data, then we need scale factors
            ReqBits +=  SelInfoToBits(LeftSelInfo[i]) + 2;  // +2 for selection bits
    }

    for (; i<sbLimit; i++) {
        if (RightBits[i] + LeftBits[i]) { //if either channel sends data, send both scale factors
            ReqBits +=  SelInfoToBits(RightSelInfo[i]) + 2; // +2 for selection bits
            ReqBits +=  SelInfoToBits(LeftSelInfo[i]) + 2;  // +2 for selection bits
        }
    }
    
	/*test code
	printf("Bound %d, req bits %d\n", bound, ReqBits);
	/*end of test code*/

    return ReqBits;
}           


int CAudioEncode::SMRToBits(int sb, float SMR)
/*
This function computes the number of bits required to encode the specified subband with
the specifed signal to mask ratio with "no signal degradation".  For "no signal degradation"
we find the smallest number of quantization bits that yields an SNR that is >= the
specified SMR.  This funciton returns the total number of bits required to encode this subband
in all 36 vectors.
*/
{
    //fetch audio allocation sequence for this subband
    const SAllocationSequence *sequence = AllocInfo->sequence[sb];

    //find the lowest SNR >= SMR
    int quants = sequence->NumQuants - 1;   //note: only check to NumQuants - 1
                                            //if no SNR is large enough we always
                                            //return the highest we can do
    if (quants < 0)   
        return 0;      //if this subband is not used, return 0 bits
    int i;
    for (i=0; i<quants; i++) {
        if (sequence->AllocationData[i]->SNR >= SMR)
            break;
    }

    return sequence->AllocationData[i]->TotalBits;

}

// insert the mnr data at the appropraite plane in the table
void CAudioEncode::insert_mnr(struct mnrrcd **mnrtab,
							  struct mnrrcd **free_list,
							  float mnr,int sbindex,int left_channel) {

	struct mnrrcd *p,*prevp,*new_entry;
	int found = 0;

	// pull an entry from the free list
	new_entry = *free_list;
	*free_list = (*free_list)->right;

	// construct the new entry
	new_entry->mnr = mnr;
	new_entry->left_channel = left_channel;
	new_entry->sbindex = sbindex;
	new_entry->right = NULL;
	new_entry->left = NULL;

	// find where to insert it
	p = *mnrtab;
	prevp = NULL;
	while (p) {
		prevp = p;
		if (mnr < p->mnr)
			p = p->left;
		else
			p = p->right;
	}

	// insert the new entry
	if (prevp == NULL)
		*mnrtab = new_entry;
	else if (mnr < prevp->mnr)
		prevp->left = new_entry;
	else
		prevp->right = new_entry;

}

void CAudioEncode::CalcBitAllocation(void)
/*
This fucction uses the current member data (StereoType, RightSMR, LeftSMR,
RightSelInfo, LeftSelInfo, JointStereoBound) to compute the bit allocation
codes (BitAllocation[]) for each subband.
*/

{
    int k,i, NewBits;
    float RightMNR[SUBBANDS];
    float LeftMNR[SUBBANDS];

	// Binary Tree used for fast sort of MNR Data
	struct mnrrcd *prevp;
	struct mnrrcd *mnrtab = NULL;
	struct mnrrcd *free_list = NULL;
	struct mnrrcd mnr_buff[SUBBANDS*2];

	//calc total bits - header
    int FreeBits = (GetNextFrameSize() - sizeof(header) ) * 8;
	
	//calc number of bits available for data, scale factors, and scale factor selection
	//by subtracting allocation bits from total
	int sb;
	int AllocBits = 0;
	for(sb=0; sb<JointStereoBound; sb++)
		AllocBits += 2*AllocInfo->sequence[sb]->NumBits;
	for(; sb<sbLimit; sb++)
		AllocBits += AllocInfo->sequence[sb]->NumBits;
    FreeBits -= AllocBits;

    //init mask to noise data  (MNR = SNR - SMR) and allocation bits
    for (i=0; i<sbLimit; i++) {     //init right data
        RightMNR[i] =  -RightSMR[i];
        RightBitAllocation[i] = 0;
    }
    for (i=0; i<JointStereoBound; i++) {    //init left data
        LeftMNR[i] =  -LeftSMR[i];
        LeftBitAllocation[i] = 0;
    }

	// construct a full free list 
	free_list = NULL;
	mnr_buff[0].right = NULL;
	for (k=0;k<SUBBANDS*2;k++) {
		prevp = free_list;
		free_list = &mnr_buff[k];
		free_list->right = prevp;
	}

	
	/* add the initial entries */
	for (k=0;k<sbLimit;k++) {
		if (k < JointStereoBound)
		    insert_mnr(&mnrtab,&free_list,LeftMNR[k],k,1);
		insert_mnr(&mnrtab,&free_list,RightMNR[k],k,0);
	}

    //add bits to the lowest mask-to-noise subband, loop until we can not
	//fit any more bits (until all subbands have been marked out)
	int sbMin,left;
    while ( (sbMin = GetMinMNR(&mnrtab,&free_list, &left)) >= 0) {

		float * MNRPtr, SMR;
		int *BitAllocPtr, SelInfo;

		if (!left) {   //min mnr is in right data
			MNRPtr = &RightMNR[sbMin];
			SMR = RightSMR[sbMin]; 
			BitAllocPtr = &RightBitAllocation[sbMin];
			SelInfo = RightSelInfo[sbMin];
		}
		else {		   //min mnr is in left data
			MNRPtr = &LeftMNR[sbMin];
			SMR = LeftSMR[sbMin]; 
			BitAllocPtr = &LeftBitAllocation[sbMin];
			SelInfo = LeftSelInfo[sbMin];
		}

        NewBits = GetAdditionalBits(sbMin, *BitAllocPtr, SelInfo);

        if(FreeBits >= NewBits) {
			//if we can fit this subband, update the allocation and number of free bits
            int NewAlloc = ++(*BitAllocPtr);
	        FreeBits -= NewBits;		//update free bits

            //update mask to noise
            if ((AllocInfo->sequence[sbMin]->NumQuants - 1)  == NewAlloc)
                //if no more allocation bits, turn off further allocation
                *MNRPtr = MAXMNR;
            else                    
                *MNRPtr = AllocToSNR(NewAlloc,sbMin) - SMR; 
        }
		else {
			//if we can't fit this subband, turn off futher allocation, and try next smallest
	        *MNRPtr = MAXMNR;
		}

		// Place the adjusted value back in the tree
		if (*MNRPtr < MAXMNR)
		    insert_mnr(&mnrtab,&free_list,*MNRPtr,sbMin,left);


    } //end while ( (sbMin = GetMinMNR(RightMNR, LeftMNR, &right)) >= 0)

	//set the left allocaton above joint stereo bound to the right allocation
	for (i=JointStereoBound; i<sbLimit; i++)
		LeftBitAllocation[i] = RightBitAllocation[i];

	/*test code
	int totbits = GetNextFrameSize()*8 - FreeBits;
	printf("frame = %3d; alloc = %3d bytes + %d bits; next alloc = %d\n",
					GetNextFrameSize(), totbits/8, totbits - (totbits/8)*8, NewBits);
	/* end of test code*/
}

/* find the record with the min mnr and remove it */
int CAudioEncode::GetMinMNR(struct mnrrcd **mnrtab,
							struct mnrrcd **free_list,
							int *left_channel) {
	struct mnrrcd *p,*prevp;

	p = *mnrtab;
	prevp = NULL;

	// empty list
	if (p == NULL)
		return(-1);

	// locate the first 
	while (p->left) {
		prevp = p;
	    p = p->left;
	}

	// Fix up the remaining tree
    if (p == *mnrtab)
		*mnrtab = p->right;
	else
		prevp->left = p->right;

	// Put the entry back in the free list
	p->right = *free_list;
	*free_list = p;

	// Setup the return
	*left_channel = p->left_channel;
	return(p->sbindex);

	
}


int CAudioEncode::GetAdditionalBits(int subband, int CurrentAllocation, int SelInfo)
/*
this function returns the number of additional bits to go to the next
allocation step for the specified subband.  The current allocation level and the
scale factor select info for the specified subband must also be supplied.
*/
{
    //ASSERT((AllocInfo->sequence[subband]->NumQuants - 1) < CurrentAllocaton)

    //compute added data bits = new total - old total
    const SAllocationSequence *sequence = AllocInfo->sequence[subband];
    int bits = sequence->AllocationData[CurrentAllocation+1]->TotalBits        //new total
                    - sequence->AllocationData[CurrentAllocation]->TotalBits;  // - old total

    //if first bits for subband, add in scale factor bits
    if (CurrentAllocation == 0) {
        int sfbits;
        if (StereoType == JOINT_STEREO && subband >= JointStereoBound) {
            //if subband in joint stereo, include both left channel and right channel scale factors
            sfbits  =  SelInfoToBits(RightSelInfo[subband]) + 2; // +2 for selection bits
            sfbits +=  SelInfoToBits(LeftSelInfo[subband]) + 2;  // +2 for selection bits
        }
        else {
            sfbits = SelInfoToBits(SelInfo) + 2;  //+2 for selection bits
        }
        bits += sfbits;
    }


    return bits;
}



void CAudioEncode::BuildOuptut(BYTE *FrameOutput)
/*
This funciton uses calculated member data to build an MPEG-1 output bit stream
*/
{
	BYTE *outptr = FrameOutput;

	//copy header
	memcpy(outptr,header,sizeof(header));

	//update  header info in output
	if (rest < 0)		//set padding bit if we are in a padding frame
		outptr[PADDING_BYTE] += PADDING_BIT;
	if (StereoType == JOINT_STEREO) {
		int mode;
		if (JointStereoBound >= sbLimit) {
			//if bound is high, send as stereo
			mode = (int)STEREO << 6;
		}
		else {
			//set d7-d6 to stereo type = joint stereo, set d5-d4 to encode joint stereo bound:
			//Bound => ((JointStereoBound - 4)/4) << 4
			mode = ((int)JOINT_STEREO << 6) + ((JointStereoBound - 4) << 2);
		}
		outptr[STEREO_BYTE] &= ~STEREO_MASK;  //zero stereo bits
		outptr[STEREO_BYTE] += mode;		  //set new stereo bits
	}

	outptr += sizeof(header);   //point to allocaton bits

	CCopyBits cb(outptr);		//create object to copy remaining bits
	int sb;

	//copy allocation bits
	for(sb=0; sb<JointStereoBound; sb++) {
		int NumBits = AllocInfo->sequence[sb]->NumBits;
		cb.CopyBits(RightBitAllocation[sb],NumBits);
		cb.CopyBits(LeftBitAllocation[sb],NumBits);
	}
	for(; sb<sbLimit; sb++) {
		int NumBits = AllocInfo->sequence[sb]->NumBits;
		cb.CopyBits(RightBitAllocation[sb],NumBits);
	}

	//copy scale factor selection info
	if (StereoType == MONO) {
		//if mono only copy right selection info
		for(sb=0; sb<sbLimit; sb++) {
			if (RightBitAllocation[sb])	 //copy if any allocatoin is set
				cb.CopyBits(RightSelInfo[sb],2);
		}
	}
	else {
		//if stereo, always copy right and left selecton info
		for(sb=0; sb<sbLimit; sb++) {
			if (RightBitAllocation[sb])	 //copy if any allocatoin is set
				cb.CopyBits(RightSelInfo[sb],2);
			if (LeftBitAllocation[sb])	 //copy if any allocatoin is set
				cb.CopyBits(LeftSelInfo[sb],2);
		}
	}

	//copy scale factor codes
	if (StereoType == MONO) {
		//if mono only copy right scale factors
		for(sb=0; sb<sbLimit; sb++) {
			if (RightBitAllocation[sb])	 //copy if any allocatoin is set
				CopyScaleFactors(cb, RightScaleFactors[sb], RightSelInfo[sb]);
		}
	}
	else {
		//if stereo, always copy right and left scale factors
		for(sb=0; sb<sbLimit; sb++) {
			if (RightBitAllocation[sb])	 //copy if any allocatoin is set
				CopyScaleFactors(cb, RightScaleFactors[sb], RightSelInfo[sb]);
			if (LeftBitAllocation[sb])	 //copy if any allocatoin is set
				CopyScaleFactors(cb, LeftScaleFactors[sb], LeftSelInfo[sb]);
		}
	}

	CopySubbandData(cb);

	cb.Flush();  //flush any uncopied bits

}

void CAudioEncode::CopySubbandData(CCopyBits &cb)
/*
This function is used to quintize and copy all of the subband data to the
specified output object.
*/
{
	int ng,gr,vec,sb;
    float InvertedScaleFactor;

	//quantize and copy stereo data
	if (StereoType != MONO) {
		for (vec=ng=0; ng<NUMGROUPS; ng++) {      //for each group
			for (gr=0; gr<GROUPSIZE; gr+=3, vec+=3) { //for each trio of vectors in the group

				for(sb=0; sb<JointStereoBound; sb++) {
					//for each subband less than joint stereo bound, copy right and left data
					const SQuantInfo * const *AllocData = AllocInfo->sequence[sb]->AllocationData;

					int alloc = RightBitAllocation[sb];
					if (alloc) { //copy if any allocatoin is set
						InvertedScaleFactor = InvertedScaleFactors[RightScaleFactors[sb][ng]];
						CopyData(cb, AllocData[alloc],
							 RightData[vec  ][sb]    * InvertedScaleFactor,
							 RightData[vec+1][sb]    * InvertedScaleFactor,
							 RightData[vec+2][sb]    * InvertedScaleFactor);
					}

					alloc = LeftBitAllocation[sb];
					if (alloc) { //copy if any allocatoin is set
						InvertedScaleFactor = InvertedScaleFactors[LeftScaleFactors[sb][ng]];
						CopyData(cb, AllocData[alloc],
							 LeftData[vec  ][sb]    * InvertedScaleFactor,
							 LeftData[vec+1][sb]    * InvertedScaleFactor,
							 LeftData[vec+2][sb]    * InvertedScaleFactor);
					}
				}
				for(; sb<sbLimit; sb++) {
					//for each subband above joint stereo bound,
					//copy (right data + left data) / (right scale factor + left scale factor)
					const SQuantInfo * const *AllocData = AllocInfo->sequence[sb]->AllocationData;
					
					int alloc = RightBitAllocation[sb];
					if (alloc) { //copy if any allocatoin is set
						InvertedScaleFactor = InvertedScaleFactors[RightScaleFactors[sb][ng]]
											+ InvertedScaleFactors[LeftScaleFactors[sb][ng]];
						CopyData(cb, AllocData[alloc],
							 (RightData[vec  ][sb] + LeftData[vec  ][sb]) * InvertedScaleFactor,
							 (RightData[vec+1][sb] + LeftData[vec+1][sb]) * InvertedScaleFactor,
							 (RightData[vec+2][sb] + LeftData[vec+2][sb]) * InvertedScaleFactor);
					}

				}
			} //end for each trio of vectors in the group
		}//end for each group
	}//end if copy stereo data

	//quantize and copy mono data
	else {
		for (vec=ng=0; ng<NUMGROUPS; ng++) {      //for each group
			for (gr=0; gr<GROUPSIZE; gr+=3, vec+=3) { //for each trio of vectors in the group

				for(sb=0; sb<sbLimit; sb++) {
					//for each subband, copy mono (right) data
					const SQuantInfo * const *AllocData = AllocInfo->sequence[sb]->AllocationData;
					
					int alloc = RightBitAllocation[sb];
					if (alloc) { //copy if any allocatoin is set
						InvertedScaleFactor = InvertedScaleFactors[RightScaleFactors[sb][ng]];
						CopyData(cb, AllocData[alloc],
							 RightData[vec  ][sb]    * InvertedScaleFactor,
							 RightData[vec+1][sb]    * InvertedScaleFactor,
							 RightData[vec+2][sb]    * InvertedScaleFactor);
					}

				}
			} //end for each trio of vectors in the group
		}//end for each group
	}//end else copy mono data

}

void CAudioEncode::CopyData (CCopyBits &cb, const CAudioEncode::SQuantInfo *QuantInfo,
							 float data0, float data1, float data2)
/*
This funciton quantizes and writes the specified pre-scaled data words to the specified output
object.
*/
{
	int NumBits = QuantInfo->CodewordBits;

	if (QuantInfo->CodewordSamps == 1) {
		//if no grouping, quantize and copy a sample form each group
		cb.CopyBits(Quantize(data0, QuantInfo), NumBits);
		cb.CopyBits(Quantize(data1, QuantInfo), NumBits);
		cb.CopyBits(Quantize(data2, QuantInfo), NumBits);
	}
	else {
		//if grouping, form into a single codeword (eg v3 = 9*z + 3*y + x)
		int out;
		int steps = QuantInfo->steps;
		out = Quantize( data2, QuantInfo);
		out = steps * out	+ Quantize( data1, QuantInfo);
		out = steps * out	+ Quantize( data0, QuantInfo);
		cb.CopyBits(out, NumBits);
	}
}

void CAudioEncode::CopyScaleFactors(CCopyBits &cb,
									int ScaleFactors[NUMGROUPS],
									int SelInfo)
/*
This function copies scale factor info from the specified subband
into the specified CopyBits object.  The funciton uses the specified
scale factor and sclae factor selecton info.
*/
{
	switch (SelInfo) {
	case 0:
		cb.CopyBits(ScaleFactors[0],6);
		cb.CopyBits(ScaleFactors[1],6);
		cb.CopyBits(ScaleFactors[2],6);
		break;
	case 1:
	case 3:
		cb.CopyBits(ScaleFactors[0],6);
		cb.CopyBits(ScaleFactors[2],6);
		break;
	case 2:
		cb.CopyBits(ScaleFactors[0],6);
		break;
	default:
		break;
	}

}
