#ifndef _ENCODE_
#define _ENCODE_

/*
encode.h
This class implements the MPEG-1 layer II audio encoding
algorithm. (see ISO/IEC 11172-3)
*/
#include "lyr2defs.h"
#include "sbfilter.h"
#include "copybits.h"
#include "delay.h"
#include "smr.h"

#define SCALEFACTORTABLESIZE 63

#define MAXMNR 1000.0F      //the max mask to noise ratio

class CAudioEncode {
    private:

    //-----static tables--------------
    static const struct SQuantInfo {
        int     steps;          //number of quantizaton steps
        int     CodewordSamps;  //samples per codeword
        int     CodewordBits;   //bits per codeword
        int     TotalBits;      //total bits for all 36 subband samples
        float   SNR;            //signal to noise in dB
        float   NormA;      //normalization factor A (prequant = A * subband / scale factor + B)
        float   NormB;      //normalization factor B (prequant = A * subband / scale factor + B)
    } QuantInfo[18];      //18 possible quantizations

    static const struct SAllocationSequence {
        int        NumBits;     //number of bits in allocation word
        int        NumQuants;   //number of possible quantizations (2^NumBits)
        const SQuantInfo *AllocationData[16];   //quant info for each possible allocation value
    } AllocSequence[7];    //7 possible allocation sequences

    static const struct SAllocationInfo {
        const SAllocationSequence *sequence[SUBBANDS];     //allocation sequence for each subband
    } AllocationInfo[4];   //4 possible allocation tables

    static const struct SAllocationSetup {
        int     rate;       //data rate per channel in kbits/second (-1 == free format)
        const SAllocationInfo *AllocInfo[3];      //look up allocation table by sample rate:
                                                        //index 0 -> 44.1 Khz
                                                        //index 1 -> 48 Khz
                                                        //index 2 -> 32 Khz
    } AllocationSetup[11];

	static const struct SBitRates {
		int rate;  		//total output data rate in Kbits/sec (-1 => free format)
		int ModeUsed;	//bit set for each allowable mode:
						//   D0=>STEREO, D1=>JOINT_STEREO, D2=>DUAL_CHANNEL, D3=>MONO
	} BitRates[15];
    
	static const float ScaleFactorTable[SCALEFACTORTABLESIZE];
	static float InvertedScaleFactors[SCALEFACTORTABLESIZE];

	static const struct SScaleFactorInfo {
		int sf0;		//group index of scale factor to use as scale factor 0 
		int sf1;		//group index of scale factor to use as scale factor 1 
		int sf2;		//group index of scale factor to use as scale factor 2
		int SelInfo;	//scale factor selection code 
	} ScaleFactorInfo[5][5]; //25 possible scale factor encoding info entries

    static const struct SScaleFactorSelInfo {
        int SFSend;      //number of scale factors sent
        int TotalBits;   //total number of scale factor bits 
    } SFSelInfo[4];      //4 possible SF selection codes

    //----data members---------------------
	BYTE header[4];				//32 bit frame header
	ESamplingFreq SamplingFreqCode;
	EStereoType StereoType;
	int Decimation;				// Decomation code for 22 and 11K sample rate
    const SAllocationInfo *AllocInfo;  //pointer to allocation info table for the current settings

	
	float RightData[VECTORS][SUBBANDS];	//an array of 36 vectors of 32 subband values
	int RightScaleFactors[SUBBANDS][NUMGROUPS];
	float RightSMR[SUBBANDS];
	int RightSelInfo[SUBBANDS];			//right scale factor selection info
    int RightBitAllocation[SUBBANDS];
	CDelay RightDelay;					//delay line for fft data used by CalcSignalToMask()

	float LeftData[VECTORS][SUBBANDS];	//an array of 36 vectors of 32 subband values
	int LeftScaleFactors[SUBBANDS][NUMGROUPS];
	float LeftSMR[SUBBANDS];
	int LeftSelInfo[SUBBANDS];			//left scale factor selection info
    int LeftBitAllocation[SUBBANDS];
	CDelay LeftDelay;					//delay line for fft data used by CalcSignalToMask()


	int BitRateCode;
	int FrameSize;   //Output frame size (in bytes)
	int dif, rest;   //frame size remainder and running remainder sum
					 //(used to determine when extra byte needed, see 11172-3 2.4.2.3)
    int SamplingFreq;//sampling frequency in Hz
    int sbLimit;     //the number of transmited subbands for this sample rate/data rate (see table B.2)
	int sbLimitVar;  // variable upper limit which may be dynamically lowered
	int sbLimitBW;   // subband limit for the input sample rate (BW)
    int JointStereoBound;         // set to sbLimit for stereo & dual channel, set to 0 for mono
	int JointSterioVoteCount; // # of votes to raise or lower bound
	int JointSterioVotes; // Vote Sum (-1 = one vote to lower; +1 = on vote to raise)

	CSubbandFilter RightFilter;
	CSubbandFilter LeftFilter;

    CCalcSMR CalcSMR;  //object used to calculate SMRs

	// Binary MNR Tree
    struct mnrrcd {
		struct mnrrcd *left;
		struct mnrrcd *right;    // Right is also used for free list pointers
		short left_channel;
		short sbindex;
		float mnr;
	};

    //----interface functions-------
    public:
    CAudioEncode(	ESamplingFreq Freq=SF_44100_HZ,
    				EStereoType Type = STEREO,
    				int BitRate = 160 /*(in KHz)*/);
	int GetNextFrameSize(void);   //returns the size of the next output frame in bytes
	int GetMaxFrameSize();        //returns the size of the largest possible output frame
	void ProcessFrame(  short *FrameInput,    //1152 16 bit samples or 2*1152 16 samples for stereo
                                              //(1152/2 samples if sample rate == 22050)
                                              //(1152/4 samples if sample rate == 11025)
						BYTE  *FrameOutput,   //(1152 * (bit rate / 8) / sample rate) bytes on 
						int   combined_smr,   //Compute SPR based on sum of both channels (faster)
						int   *MaxSB);        // MAX Subband to process (0= use default)
											  // default returned

											  //average (output size may vary by 1 byte)
	void ProcessFrame(  float (*FrameInput)[2], //1152 stereo floating point input samples
						BYTE  *FrameOutput,   //(1152 * (bit rate / 8) / sample rate) bytes on 
											  //average (output size may vary by 1 byte)
						int   combined_smr,   //Compute SPR based on sum of both channels (faster)
						int   *MaxSB);        // MAX Subband to process (0= use default)
											  // default returned

	// Note: The processing speed may be adjusted by raising or lowering MaxSB to
	//       conform the the avaialble processing time. Lowering MaxSB < 22 will result
	//       in distorting the input signal by eliminating some of the high frequency 
	//       components of the signal.
											  

	
	//----helper functions-----------
	private:
	//compute scale factor indexes and selection info form subband data
	void CalcScaleFactors(float SubBandData[VECTORS][SUBBANDS],
						int ScaleFactors[SUBBANDS][NUMGROUPS],
						int SelInfo[SUBBANDS]);
	
	void ProcessCurData(int combined_smr,BYTE *FrameOutput);
	int CAudioEncode::IncrementBound (int current_bound);
	int CAudioEncode::DecrementBound (int current_bound);
	int CalcJointStereoBound(void);
    int GetRequiredBits(int bound, int RightBits[SUBBANDS], int LeftBits[SUBBANDS]);
	int CalcScaleFactorClass(int sf1, int sf2);
    int SMRToBits(int sb, float SMR);
    void CalcBitAllocation(void);
	int GetMinMNR(struct mnrrcd **mnrtab,struct mnrrcd **free_list,
		          int *left_channel);
	void insert_mnr(struct mnrrcd **mnrtab,
					struct mnrrcd **free_list,
					float mnr,int sbindex,int left_channel);
    int GetAdditionalBits(int subband, int CurrentAllocation, int SelInfo);
    void BuildOuptut(BYTE *FrameOutput);
    void CopyScaleFactors(CCopyBits &cb, int ScaleFactors[NUMGROUPS], int SelInfo);
	void CAudioEncode::CopySubbandData(CCopyBits &cb);
	void CopyData(CCopyBits &cb, const CAudioEncode::SQuantInfo *QuantInfo,
							 float data0, float data1, float data2);

    
    
    //inline helper functions

    int SelInfoToBits(int SelInfo)
    /*
    This function returns the number of bits needed for the scale factor based on the 
    specified scale factor selecton code
    */
    {
        //ASSERT(SelInfo >= 0);
        //ASSERT(SelInfo < 4);
        return SFSelInfo[SelInfo].TotalBits;
    }


    float AllocToSNR(int AllocCode,int subband)
    /*
    this function returns the signal to noise ratio of the specified subband with the
    specifed bit allocation code.
    */
    {
        return AllocInfo->sequence[subband]->AllocationData[AllocCode]->SNR;
    }

	int Quantize (float data, const CAudioEncode::SQuantInfo *QInfo)
	/*
	This function quantizes the specified input based on the specified
	QuantInfo structure.  The function returns the quantized bits as an integer.

    note that NormA and NormB are scaled so that qauntized bits are to the left
    of the binary point.  Also, NormB includes the offset needed to provide
    the required conversion to offset binary output.  This allows conversion
    before truncation so that the output will be truncated down for both
    positive and negative input
	*/
	{
		return (int)(QInfo->NormA*data + QInfo->NormB);
	}

    //absolute value
    float absval(float in) {return in < 0 ? -in : in;} 
};


#endif //#ifndef _ENCODE_
