/* smr.c
This file contains an implementation of the psychoacoustic model 1 form 
ISO/IEC 11172-3 Annex D.  This code is the implementation of the CCalcSMR class.
*/

#include "smr.h"
#include "fft.h"
#include <math.h>

//#define MATCH_MPEG92_CODE

/*
 * Fast liner to db calculator tables
 */
struct CCalcSMR::logtabrcd logtab[LOGTABSIZE];
struct CCalcSMR::logtabrcd *logtabzero;
float seven_dBLin;
float AbsThreshOffsetLin;

// Absolute threshold converted to a linear fraction
float CCalcSMR::AbsThrsLin[132];


CCalcSMR::CCalcSMR(ESamplingFreq Freq/*=SF_44100_HZ*/, int BitRate/*=160(in KHz)*/)
/*
class constructor:  init tables, pointers and data based on sampling freq and bitrate
*/
{
	// Threshold used in isTone
	seven_dBLin = (float)pow(10.0F, 0.1F * 7.0F);

	

	// Init the lobtab for dBToPower
    int k;
	for (k=0;k<LOGTABSIZE;k++) {
		float x = MINLOG + (float)(k);
		logtab[k].value = (float)pow(10.0F, 0.1F * x);
		float val1 = (float)pow(10.0F, 0.1F * (x + 1.0F/10));
		logtab[k].slope = (val1 - logtab[k].value) / (1.0F/10);
	}

	// use this pointer to avoid the subtract
	logtabzero = &logtab[LOGTABOFZERO];
			               
    float BinWidth;  //bandwidth of fft bin in Hz

    if (BitRate >= 96)
        AbsThreshOffset = -12.0F; //use lower absolute threshold if output rate is reasonalbe
    else
        AbsThreshOffset = 0.0F;
	AbsThreshOffsetLin = (float)pow(10.0F, 0.1F * AbsThreshOffset);

    //init tables d1 and d2
    switch (Freq) {
    case SF_32000_HZ:
        SubSamp = SubSamp32000;
        SubSampSize = sizeof(SubSamp32000)/sizeof(SubSamp32000[0]);
        CritBands = CritBands32000;
        NumCritBands = sizeof(CritBands32000)/sizeof(CritBands32000[0]);
        BinWidth = 32000.0F/FFTSIZE;
        break;
    case SF_44100_HZ:
	case SF_22050_HZ:
	case SF_11025_HZ:
        SubSamp = SubSamp44100;
        SubSampSize = sizeof(SubSamp44100)/sizeof(SubSamp44100[0]);
        CritBands = CritBands44100;
        NumCritBands = sizeof(CritBands44100)/sizeof(CritBands44100[0]);
        BinWidth = 44100.0F/FFTSIZE;
        break;
    case SF_48000_HZ:
        SubSamp = SubSamp48000;
        SubSampSize = sizeof(SubSamp48000)/sizeof(SubSamp48000[0]);
        CritBands = CritBands48000;
        NumCritBands = sizeof(CritBands48000)/sizeof(CritBands48000[0]);
        BinWidth = 48000.0F/FFTSIZE;
        break;
    default:
        //ASSERT(0);
        break;
    }

	// Compute the SubSamp liner thershold values
	for (k=0;k<SubSampSize;k++)
		AbsThrsLin[k] = (float)pow(10.0F, 0.1F * SubSamp[k].AbsThrs);


    //compute BinToIndex table entries
    int bin;
    int index = 0;
    for (bin=0; bin<FFTSIZE/2 && (index+1) < SubSampSize; bin++) {
        if (abs(bin-SubSamp[index].fftBin) > abs(bin-SubSamp[index+1].fftBin) ) {
            index++; //if bin freq is closer to next sub-sampled freq, inc index
        }
        BinToIndex[bin] = index;
    }
    for (; bin<FFTSIZE/2; bin++) {
        //if we are at end of sub-sampled table, fill remaining bins with highest values
        BinToIndex[bin] = SubSampSize - 1;
    }

    //init fft & window

#ifdef MATCH_MPEG92_CODE
	//constants to match the MPEG-92 test code
	WINDOWTYPE WinType = HANN;		//don't normalize window
	float WinScale = 1.0F/(float)FFTSIZE;  //scale for 90 db
#else
	//constants to scale FFT to max of 
	WINDOWTYPE WinType = HANNING;		//normalize window
	float WinScale = 2.0F/(float)FFTSIZE;  //scale for 96 db
#endif

    InitFFT();
    BuildWindow(WinType, FFTSIZE, HannWindow);	//use HANNING window because it is normalized

    //normalize window by fft size so max fft output is 96dB
    int i;
    for (i=0; i<FFTSIZE; i++)
        HannWindow[i] *= WinScale;
}





void CCalcSMR::CalcSignalToMask(
			float RawData[INPUTSIZE],
			int ScaleFactors[SUBBANDS][NUMGROUPS],
			float SMR[SUBBANDS], int ValidSubbands)
/*
compute the signal to mask ratio (SMR) for the specified number of valid subbands
from the specified raw input (already delayed to match the sbband filter)
with the specifed scale factors.  The results are written to the specified SMR array.
*/
{
    int i,j;

	//window and FFT
    for (i=0; i<FFTSIZE; i++)
        fftPow[i] = RawData[i] * HannWindow[i];
    DoRealFFT(fftPow, fftPow, NULL, NULL, LOG2FFTSIZE);

    //compute signal power
    CalcSubbandPower(ScaleFactors, ValidSubbands);
    
    //find frequency components
    NumTones = FindTones(Tones);
    NumNonTones = FindNonTones(NonTones);

    //------ find the global mask threshold for each sub-sampled frequency -----

    //first compute threshold of quiet contribution
    for (i = 0; i < SubSampSize; i++)
        GlobalMaskThold[i] = AbsThrsLin[i] * AbsThreshOffsetLin;

    //second compute cntribution from each tonal components
    for (j=0; j<NumTones; j++)
        UpdateGlobalMasking(Tones[j], 1);

    //third compute contribution from non-tonal components
    for (j=0; j<NumNonTones; j++)
        UpdateGlobalMasking(NonTones[j], 0);

    //---------------------------------------------------------------------------


    //calculate the SMR for each subband
    int bin;           //fft bin
    int sb;            //subband
    int index = 0;     //sub-sampled frequency index
    for (sb=0, bin=BINSPERSUBBAND; sb<ValidSubbands; sb++, bin+=BINSPERSUBBAND) {
        float mask;

        //first find min global masking threshold
        if (index < SubSampSize) {
            //check each sub-sample in subband
            mask = GlobalMaskThold[index++]; 
            while (index < SubSampSize && SubSamp[index].fftBin < bin)
                mask = min(mask, GlobalMaskThold[index++]);
            mask = PowerTodB(mask);   //don't forget to convert to dB
        }
        else {
            //subband is beyond end of sub-sample table,
            //just use threshold of quiet of last entry in table
            mask = SubSamp[SubSampSize-1].AbsThrs;
        }

        //finally, lets compute the signal to mask ratio (in dB)
        SMR[sb] = SubbandPower[sb] - mask;
        
    }
}

void CCalcSMR::CalcSubbandPower(int ScaleFactors[SUBBANDS][NUMGROUPS], int ValidSubbands)
/*
This function computes the power level of each valid subband based on the fft
member data and the specified scale factros.   The function updates the SubbandPower[]
memeber data.
*/
{
    int sb,j;
    int bin = 0;       //fft bin
    int index = 0;     //sub-sampled frequency index
    for (sb=0; sb<ValidSubbands; sb++) {

        //first compute the total subband power from fft
        SubbandPower[sb] = fftPow[bin++];
        for (j=1; j<BINSPERSUBBAND; j++)
            SubbandPower[sb] += fftPow[bin++];   //compute total power
        SubbandPower[sb] = PowerTodB(SubbandPower[sb]);  //convert to dB

        //next, check scalefactor power
        int minindex = min( ScaleFactors[sb][0], min(ScaleFactors[sb][1],ScaleFactors[sb][2]) );
        float ScaleFactorPower = dBScaleFactorTable[minindex];  //max scalefactor at min index

        //use larger of scalefactror or fft power
        SubbandPower[sb] = max(SubbandPower[sb],ScaleFactorPower);
    }
}


int CCalcSMR::FindTones(SSignalComponents component[MAXTONES])
/*
This funciton uses the current member data to compute the number of tonal
components in the signal.  It returns the number of tones found and fills in
the component data structure for each tone.
*/
{
    int ToneCount = 0;
    int bin;
    float LastCritBandRate = -1.0F;

    //for each bin that could be a tone
	for (bin=2; bin<(FFTSIZE/2 - 12); bin++) {
        float BinMag = fftPow[bin];

        //if we have a local maximum
		if (BinMag > fftPow[bin-1] && BinMag > fftPow[bin+1] ) {
            
            int ToneWidth = BinToWidth(bin);

			//if we have a new tone
            if (IsTone(bin, ToneWidth) ) {

				//ok we found a tone, 
				//Are we at least .5 bark from last tone?
				if ((SubSamp[BinToIndex[bin]].CritBandRate - LastCritBandRate) >= .5F) {
					//we are >=	.5 bark away, add new tone to list
					StoreToneData(&component[ToneCount], bin);
					LastCritBandRate = component[ToneCount].CritBandRate;
					ToneCount++;
					//ASSERT(ToneCount < MAXTONES);
				}
				else if (BinMag > dBToPower(component[ToneCount-1].MagdB)) {
					//we are within .5 bark, but the current tone is larger
					//so overwite old tone
					StoreToneData(&component[ToneCount-1], bin);
					LastCritBandRate = component[ToneCount-1].CritBandRate;
				}

				//let's zero out fft power data
				//to prepare for non-tonal computation
				int j;
				int ToneEnd = bin+ToneWidth;
				for (j = bin-ToneWidth; j <= ToneEnd; j++)
					fftPow[j] = 0.0F;


			} //end if we have a new tone
        } //end if we have a local maximum
    }  //end for each bin that could be a tone

    return ToneCount;
}

void CCalcSMR::StoreToneData(SSignalComponents *component, int bin)
/*
This function fills in the specifed component data structure with tone
info from the specifed bin
*/
{
	int SubSampIndex = BinToIndex[bin];
	component->SubSampIndex = SubSampIndex;
	component->CritBandRate = SubSamp[SubSampIndex].CritBandRate;
	component->MagdBLin = fftPow[bin-1] + fftPow[bin] + fftPow[bin+1];
	component->MagdB = PowerTodB(component->MagdBLin);
}            



int CCalcSMR::BinToWidth(int bin)
/*
find required tone width (in bins) based on bin.  In this case width is defined to be
form one edge of the tone to the center bin.
*/
{
    int width;
    if (bin >= 255)
        width = 12;
    else if (bin >= 127)
        width = 6;
    else if (bin >= 63)
        width = 3;
    else
        width = 2;
    return width;
}


int CCalcSMR::IsTone(int bin, int ToneWidth)
/*
This function retruns TRUE (1) if there is a tone at the specified bin, FALSE (0)
otherwise.  The function checks that the tone is > 7 db larger then it's neighboring
bins and that it is above the absolute threshold
*/
{
    float BinMag = fftPow[bin];

    //make sure that tone is at least 7 db larger than any neighboring bin
    int i;
    for (i=2; i<=ToneWidth; i++) {
        if ( (BinMag / fftPow[bin-i]) < seven_dBLin || (BinMag /  fftPow[bin+i]) < seven_dBLin)
            return 0;
    }

    //check against threshold in quite (absolute threshold)
    if (BinMag < AbsThrsLin[BinToIndex[bin]] * AbsThreshOffsetLin)
        return 0;


    return 1;
}


int CCalcSMR::FindNonTones(SSignalComponents component[MAXNONTONES])
/*
This funciton uses the current member data to compute the number of non-tonal
components in the signal.  It returns the number of non-tones found and fills in
the component data structure for each non-tone.
*/
{
    int bin = 1;  //start at bin 1, skip DC term
	int band;

	//for each critical band
	for (band=0; band < NumCritBands; band++) {

		//find total power
		float totpow = 0.0F;
		int MaxBin = CritBands[band].MaxBin;
		while (bin <= MaxBin)
			totpow += fftPow[bin++];
			
		//store non-tonal component info
		int SubSampIndex = BinToIndex[CritBands[band].MidBin];
		component[band].SubSampIndex = SubSampIndex;
		component[band].CritBandRate = SubSamp[SubSampIndex].CritBandRate;
		component[band].MagdBLin = totpow;
		component[band].MagdB = PowerTodB(totpow);
	}

	return NumCritBands;
}


#include <stdio.h>


void CCalcSMR::UpdateGlobalMasking(SSignalComponents &comp, int IsTone)
/*
this function updates the global masking threshold for each subsampled frequency
by adding the masking contribution from the specified component.  The IsTone parameter
is 1 if the component is a tone, 0 if the component is noise 
*/
{
    float dz;  //bark difference
	float BaseMask;
    float mask;

    if (IsTone)
		BaseMask = comp.MagdB  - (4.5F + 1.525F) - 0.275F * comp.CritBandRate;
	else
		BaseMask = comp.MagdB  - (0.5F + 1.525F) - 0.175F * comp.CritBandRate;

	
	//update subsampled freq at 0 bark diff
	int SubSampIndex = comp.SubSampIndex;
	GlobalMaskThold[SubSampIndex] += dBToPower(BaseMask);

  	//update subsampled freqs at 0 < bark diff < 1
	if (++SubSampIndex >= SubSampSize)
		dz = -10.0F;			//don't check if we are at end of subsamp freq array
	else
		dz = SubSamp[SubSampIndex].CritBandRate - comp.CritBandRate;

	while (dz < 1) {
        mask = BaseMask + -17.0F * dz;
		GlobalMaskThold[SubSampIndex] += dBToPower(mask);
		if (++SubSampIndex >= SubSampSize) {
			dz = -10.0F;
			break;			//stop checking if we are at end of subsamp freq array
		}
		dz = SubSamp[SubSampIndex].CritBandRate - comp.CritBandRate;
	}

	//update subsampled freqs at 1 <= bark diff < 8
	while (dz < 8) {
        mask = BaseMask +  (1.0F - dz) * (17.0F - .15F * comp.MagdB) - 17.0F;
		GlobalMaskThold[SubSampIndex] += dBToPower(mask);
		if (++SubSampIndex >= SubSampSize)
			break;			//stop checking if we are at end of subsamp freq array
		dz = SubSamp[SubSampIndex].CritBandRate - comp.CritBandRate;
	}

	//update subsampled freqs at 0 > bark diff >= -1
	SubSampIndex = comp.SubSampIndex - 1;
	if (SubSampIndex < 0)
		dz = -5.0F;		//stop checking if we are at start of subsamp freq array
	else
		dz = SubSamp[SubSampIndex].CritBandRate - comp.CritBandRate;

	while (dz >= -1) {
        mask = BaseMask + (.4F * comp.MagdB + 6.0F ) * dz;
		GlobalMaskThold[SubSampIndex] += dBToPower(mask);
		if (--SubSampIndex < 0) {
			dz = -5.0F;
			break;			//stop checking if we are at start of subsamp freq array
		}
		dz = SubSamp[SubSampIndex].CritBandRate - comp.CritBandRate;
	}

	//update subsampled freqs at -1 > bark diff >= -3
	while (dz >= -3) {
        mask = BaseMask + 17.0F * (dz + 1.0F) - (.4F * comp.MagdB + 6.0F);
		GlobalMaskThold[SubSampIndex] += dBToPower(mask);
		if (--SubSampIndex < 0)
			break;			//stop checking if we are at start of subsamp freq array
		dz = SubSamp[SubSampIndex].CritBandRate - comp.CritBandRate;
	}

}

