#include "stream.h"
#include "..\errors.h"

//table to convert bit rate index to bitrate
int BitRateTable[16] = {
    0, 32000, 48000, 56000, 64000, 80000, 96000, 112000,
    128000, 160000, 192000, 224000, 256000, 320000, 384000, 0
};

int CStreamMixAudioPin::MoveData(char *OutPtr, int OutSize)
/*
This function is used to move the specified number of bytes to the 
specified memory.  The function returns the number of bytes moved.
*/
{
    int BytesMoved = 0;

    while (BytesMoved < OutSize) {
        if (mBytesToCopy == 0) {
            //get another full buffer if needed
            mCurrentBuff = mFull.get();
            if (mCurrentBuff == NULL)
                break;  //exit loop if no more data

            //update data queue size
        	mCritSec.Lock();
            mFullQueueBytes -= mCurrentBuff->ValidBytes;
	        mCritSec.Unlock();
            //update buffer into
            mBytesToCopy = mCurrentBuff->ValidBytes;
            mCurrentBuffPtr = mCurrentBuff->data;
        }
        int sz = MIN(OutSize-BytesMoved,mBytesToCopy);
        memcpy(OutPtr, mCurrentBuffPtr, sz);
        OutPtr += sz;
        mCurrentBuffPtr += sz;
        BytesMoved += sz;
        mBytesToCopy -= sz;
        if (mBytesToCopy == 0) {
            //if were done with the current buffer, put it on empty queue
            mEmpty.put(mCurrentBuff);
            ReleaseSemaphore(mEmptyBuffSemaphore, 1, NULL);  //signal 1 more buffer ready
        }

        //update frame counters
        mBytesRemaining -= sz;
        while ((int)mBytesRemaining < 0) {
            mFrameCount++;
            mBytesRemaining += mAveFrameSize;
        }

    }

    return BytesMoved;

}

void CStreamMixAudioPin::ProcessData(void)
/*
This function is called to process input data.
*/
{
    mOwner->ProcessData();
}

void CStreamMixAudioPin::PutBuffer(SBuff * buff)
/*
update valid data count and queue buffer
*/
{
    if (!mHeaderReady) {
        //if 1st buffer, get info from header
        ReadHeader(buff);
        mHeaderReady = TRUE;
    }

	mCritSec.Lock();
    mFullQueueBytes += buff->ValidBytes;
	mCritSec.Unlock();

    CBasePin::PutBuffer(buff);
}

void CStreamMixAudioPin::ReadHeader(SBuff * buff)
/*
This funciton extracts frame rate and audio data rate information
form the header in the specified audio buffer.  These values are stored
in the appropriate member data.
*/
{
    //search for syncword
    unsigned short sync = (unsigned short)0xfff0;
    unsigned short acc = (unsigned short)0;
    int SearchSize = buff->ValidBytes-1;
    int i;
    for (i=0; i<SearchSize; i++) {
        acc = (acc << 8) + (unsigned char)buff->data[i];
        if ((acc & (unsigned short)0xfff0) == sync)
            break;
    }
    //flag error if sync not found
    if (i == SearchSize) {
        ProcessError(ERROR_INVALID_AUDIO_DATA);
        return;
    }

    //read layer to compute audio presentation unit size
    int LayerCode = (buff->data[i] >> 1) & 0x03;
    int DecodedFrameSize;
    switch (LayerCode) {
    case 0:
        ProcessError(ERROR_INVALID_AUDIO_DATA);  
        return;
    case 1:
        ProcessError(ERROR_LAYER_3_NOT_SUPPORTED); //layer III
        return;
    case 2:
        DecodedFrameSize = 36*32;    //layer II
        break;
    case 3:
        DecodedFrameSize = 12*32;    //layer I
        break;
    }
        

    //point to 1st full byte after audio sync
    i++;

    //read bitrate index and compute data rate in bytes/second
    int BitRateIndex = (buff->data[i] >> 4) & 0x0f;

    if (BitRateIndex == 0) {
        ProcessError(ERROR_FREE_AUDIO_RATE_NOT_SUPPORTED);
        return;
    }
    if (BitRateIndex == 15) {
        ProcessError(ERROR_INVALID_AUDIO_DATA);
        return;
    }

    mDataRate = BitRateTable[BitRateIndex]/8;

    //read sample rate, and compute frame rate
    int SampleRateCode = (buff->data[i] >> 2) & 0x03;
    double SampleRate;
    switch (SampleRateCode) {
    case 0:
        SampleRate = 44100.0;
        break;
    case 1:
        SampleRate = 48000.0;
        break;
    case 2:
        SampleRate = 32000.0;
        break;
    case 3:
        ProcessError(ERROR_INVALID_AUDIO_DATA);  
        return;
    }
    mFrameRate = SampleRate/DecodedFrameSize;
    mAveFrameSize = mDataRate / mFrameRate;    

}


void CStreamMixAudioPin::InputFinished(void)
/*
This function is called when all input data is finished.
It flags this state and calls the owning object to process data
*/
{
        mInputFinished = TRUE;
        mOwner->ProcessData();
};

