/*
 * vtdisp.c
 *
 *
 * Video Tool WinNT kernel-mode driver 
 * Main I/O dispatch support
 *
 * Mark Vrabel, June 96
 */
                                 
#include "vtkernel.h"		


/*
 * 
 * VT_Dispatch
 * 
 *   Handle IRP requests and distribute them to the 
 *   appropriate functions.
 * 
 * Arguments
 * 
 *   pDeviceObject - object for target device
 *   pIrp - request packet to be handled
 * 
 * Return Value
 *
 *   Final function status.
 */
NTSTATUS
VT_Dispatch(
    IN  PDEVICE_OBJECT pDeviceObject,
    IN  PIRP pIrp
)
{

    PIO_STACK_LOCATION pIoStack;
    NTSTATUS Status;
    PDEVICE_EXT pDevExt;
    ULONG IoCode;

    /* get the device info in the extension */
    pDevExt = (PDEVICE_EXT) pDeviceObject->DeviceExtension;

    pIoStack = IoGetCurrentIrpStackLocation(pIrp);
    pIrp->IoStatus.Information = 0;

    switch(pIoStack->MajorFunction) {

	    case IRP_MJ_CREATE:

			/* 
			 * map the device information required by user
			 * mode ito this process's address space.  If 
			 * the device info is already mapped in to this
			 * process, the function will just increment the 
			 * usage count.
			 */
			if (pDevExt->pUserInfo == NULL) {
				KdPrint(("Device Opened - With no resources"));
				Status = STATUS_SUCCESS;
			} else {
				Status = VT_MapDevInfoToUser(pDevExt);
			}	
			if (NT_SUCCESS(Status)) {
				/* 
				 * just for kicks, keep track of how many times 
				 * the device is opened.
				 */
				pDevExt->DeviceInUse++;
				KdPrint(("Device Opened - DeviceInUse: %d", pDevExt->DeviceInUse));
			}
			break;

		case IRP_MJ_CLOSE:

			/* decrement the device usage count */
			--pDevExt->DeviceInUse;
			
			/* unmap the device info from this processes address 
			 * space. If the mapping is still in use by another 
			 * instance within the same process, this function 
			 * will just decrement the process usage count.
			 */
			if (pDevExt->pUserInfo == NULL) {
				KdPrint(("Device Closed - With no resources"));
			} else {
				VT_UnMapDevInfo(pDevExt);
			}

			KdPrint(("Device Closed - DeviceInUse: %d", pDevExt->DeviceInUse));
			Status = STATUS_SUCCESS;
			break;	

		case IRP_MJ_DEVICE_CONTROL:

			/*
			 * in all cases:
			 *  -check that the buffer size is big enough to hold the
			 *   input or output parameters.
		     *
 		     *  -check that the h/w-specific code has installed a
		     *   handler for this function.
		     */

			IoCode = pIoStack->Parameters.DeviceIoControl.IoControlCode;

			switch(IoCode) {

				case IOCTL_VIDTOOL_GET_DEVICEINFO:

					/* 
					 * verify the input buffer length - The DevNode is for Win95 
					 * compatability, where a specific board is identified by the
					 * DevNode and not by a named link.
					 */
					if (pIoStack->Parameters.DeviceIoControl.InputBufferLength < sizeof(DEVNODE)) {
						Status = STATUS_BUFFER_TOO_SMALL;
						break;
					} else {

						if (pIoStack->Parameters.DeviceIoControl.OutputBufferLength < sizeof(DEVICE_INFO)) {
							Status = STATUS_BUFFER_TOO_SMALL;
							break;
						} else {

							PDEVICE_INFO pDevInfo = NULL;

							/* get the device info pointer */
							pDevInfo = (PDEVICE_INFO)pIrp->AssociatedIrp.SystemBuffer;

							/* 
							 * fill in the device info structure with the device 
							 * addressed mapped into the current process space
							 */
							if (pDevExt->pUserInfo == NULL) {
								Status = STATUS_INSUFFICIENT_RESOURCES;
							} else {
								Status = VT_GetDevInfoForUser(pDevExt, pDevInfo);
							}
							
							if (NT_SUCCESS(Status)) {
								pIrp->IoStatus.Information = sizeof(DEVICE_INFO);
							}
						}
					}
					break;

				default:
					KdPrint(("Unsupported Ioctl %d", IoCode));
					Status = STATUS_INVALID_DEVICE_REQUEST;
					break;
			}
	
			break;

		default:

			KdPrint(("Unsupported major function %d", pIoStack->MajorFunction));
			Status = STATUS_INVALID_DEVICE_REQUEST;
			break;
	}

    /* 
	 * fill in the status of the IRP, and complete it unless it
     * has been queued.
     */
    if (Status != STATUS_PENDING) {
		pIrp->IoStatus.Status = Status;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
    }

    return(Status);
}


