/*
 * init.c
 *
 *
 * 32-bit Video Tool driver
 * hardware specific initialization functions 
 * for Data Translation Video Tool.
 *
 * Mark Vrabel, October 95
 */

#include <vcdriver.h>
#include <vtdioc.h>
#include <stdio.h>

#define  _INC_MMTIMERS_CODE_ TRUE
#include <mmtimers.h>

#include "hardware.h"
#include "bwuser.h"
#include <bwdlgs.h>
#include <math.h>
#include <bwhelper.h>


#define NUM_HEADERS		 (5)
#define CAL_TIME_SECONDS (12)
#define MAX_CAL_TRIES    (7)
#define CAL_FACTOR		 ((double)0.000009)
#define TIMEOUT_COUNT    (0x100000)

#define MAX_PATHNAME 256
#define MAX_REGISTRY_PATH 256


/* forward declaration of functions */
BOOL HW_Init(VCUSER_HANDLE vh, LPTSTR lpDriverName, DWORD DeviceIndex);

/* global hardware pointer for the current process */
LPHWINFO glpHwCurrent = NULL;
    

/* --- Device load-time initialization -------------------------------*/

VCUSER_HANDLE
DeviceEntry(LPTSTR lpDriverName, DWORD DeviceIndex)
{
    VCUSER_HANDLE vh;
    LPHWINFO lpHw;
    LPVC_CALLBACK lpCallback;

    /*
     * initialize the device and allocate device-extension data
     */
    vh = VC_Init(sizeof(HWINFO));

    if (vh == NULL) {
        return(NULL);
    }

    /* initialize the hardware info structure */
    lpHw = (LPHWINFO)VC_GetHWInfo(vh);
    FillMemory (lpHw, sizeof(HWINFO), 0);
    lpHw->SizeStruct = sizeof(HWINFO);

    /*
     * fill in the callbacks before hw_init 
     */
    lpCallback = VC_GetCallbackTable(vh);

    lpCallback->ConfigFormatFunc = HW_ConfigFormat;
    lpCallback->ConfigSourceFunc = HW_ConfigSource;
    lpCallback->ConfigDisplayFunc = HW_ConfigDisplay;

    lpCallback->GetOverlayModeFunc = HW_GetOverlayMode;
    lpCallback->SetKeyRGBFunc = HW_SetKeyRGB;
    lpCallback->SetKeyPalIdxFunc = HW_SetKeyPalIdx;
    lpCallback->GetKeyColourFunc = HW_GetKeyColour;
    lpCallback->SetOverlayWindowFunc = HW_SetOverlayWindow;
    lpCallback->SetOverlayRectsFunc = HW_SetOverlayRects;
    lpCallback->SetOverlayOffsetFunc = HW_SetOverlayOffset;
    lpCallback->OverlayFunc = HW_Overlay;
    lpCallback->CaptureFunc = HW_Capture;
    lpCallback->DrawFrameFunc = HW_DrawFrame;
    lpCallback->FrameFunc = HW_Frame;

    lpCallback->StreamInitFunc = HW_StreamInit;
    lpCallback->StreamFiniFunc = HW_StreamFini;
    lpCallback->StreamStartFunc = HW_StreamStart;
    lpCallback->StreamStopFunc = HW_StreamStop;
    lpCallback->StreamGetPositionFunc = HW_StreamGetPosition;

    lpCallback->RepeatFrame = HW_RepeatFrame;
    lpCallback->InterruptAcknowledge = HW_InterruptAcknowledge;
    lpCallback->CaptureService = HW_CaptureService;

	lpCallback->StreamRemoveReadyHeadFunc = NULL;
	lpCallback->StreamAddReadyTailFunc = NULL;
	lpCallback->StreamReplaceReadyHeadFunc = NULL;

	lpCallback->StreamReportSkipFunc = HW_StreamReportSkip;

    lpCallback->CleanupFunc = HW_Cleanup;
    lpCallback->DeviceOpenFunc = HW_DeviceOpen;

    /*
     * call the last-close function on cleanup as well as last-close.
     * - whether or not we are going to be re-opened, we need to
     * shut the card down.
     */
    lpCallback->DeviceCloseFunc = HW_Cleanup;

    if (!HW_Init(vh, lpDriverName, DeviceIndex)) {
        VC_Cleanup(vh);
        return(NULL);
    }

    return (vh);
}


/*
 * local audio/video synchronization calibration
 */ 
BOOL HW_CalibrateAudio (HWND hWnd, LPCSTR lpText, double * lpdfCalFactor)
{
	HWAVEIN hWave;
    WAVEFORMATEX Format;
    WAVEHDR waveHdr[NUM_HEADERS];
	BOOL bCaptureOK = FALSE;
	int Index;
    
    DWORD dwTime, dwCount = 0;
    DWORD cbLast = 0, dwBytes = 0;
    DWORD cbStart = 0, dwStart = 0;
	DWORD dwPrevious = 0;
	MMTIME mmTime;
	LPVOID lpPb = NULL;

	double dfDiff=0.0, dfCalc;
	DWORD dwRecordBytes;

	UINT wBitsPerSample = 16; 
	DWORD nSamplesPerSec, to;
	UINT nChannels; 
    WAVEINCAPS wic;
	DWORD dwCapTime = CAL_TIME_SECONDS;

	/* if hosed return default */
	*lpdfCalFactor = 1.0;

	/* verify we have a wave in device */
	if (waveInGetNumDevs() == 0) {
        return FALSE;
    }

	/* get supported format */
    if (waveInGetDevCaps(WAVE_MAPPER, &wic, sizeof(WAVEINCAPS)) != 0) {
        return FALSE;
	}

	/* make sure it supports at least one broadway format */
	if (wic.dwFormats & WAVE_FORMAT_4S16) {
		nChannels = 2; 
		nSamplesPerSec = AUDIO_RATE_44KHZ;
	} else if (wic.dwFormats & WAVE_FORMAT_4M16) {
		nChannels = 1; 
		nSamplesPerSec = AUDIO_RATE_44KHZ;
	} else if (wic.dwFormats & WAVE_FORMAT_2S16) {
		nChannels = 2; 
		nSamplesPerSec = AUDIO_RATE_22KHZ;
	} else if (wic.dwFormats & WAVE_FORMAT_2M16) {
		nChannels = 1; 
		nSamplesPerSec = AUDIO_RATE_22KHZ;
	} else if (wic.dwFormats & WAVE_FORMAT_1S16) {
		nChannels = 2; 
		nSamplesPerSec = AUDIO_RATE_11KHZ;
	} else if (wic.dwFormats & WAVE_FORMAT_1M16) {
		nChannels = 1; 
		nSamplesPerSec = AUDIO_RATE_11KHZ;
	} else {
        return FALSE;
	}
            
	/* init the wave headers */
	for (Index=0;Index<NUM_HEADERS;Index++) {
		waveHdr[Index].lpData		   = NULL;
		waveHdr[Index].dwBufferLength  = 0x10000;
		waveHdr[Index].dwBytesRecorded = 0;
		waveHdr[Index].dwUser          = 0;
		waveHdr[Index].dwFlags         = 0;
		waveHdr[Index].dwLoops         = 0;
		waveHdr[Index].lpNext          = 0;
		waveHdr[Index].reserved        = 0;
	}
					 
    /* set up the requested audio mode */
    Format.wFormatTag      = WAVE_FORMAT_PCM;        
    Format.nChannels       = nChannels;         
    Format.wBitsPerSample  = wBitsPerSample;    
    Format.nSamplesPerSec  = nSamplesPerSec;   
    Format.nBlockAlign     = (Format.nChannels * Format.wBitsPerSample) / 8;      
    Format.nAvgBytesPerSec = Format.nSamplesPerSec * Format.nBlockAlign;  
    Format.cbSize          = 0;            

	/* set the amount of bytes to record */
	dwRecordBytes = Format.nAvgBytesPerSec * dwCapTime;

	/* open the wave input device */
    if (waveInOpen(&hWave, WAVE_MAPPER, &Format, 0, 0, 0) != 0) {
        return FALSE;
    }

    /* allocate the wave buffers */
	for (Index=0;Index<NUM_HEADERS;Index++) {
		waveHdr[Index].lpData = (LPSTR)GlobalAllocPtr (GPTR, 0x10000);
		if (waveHdr[Index].lpData == NULL) {
			goto clean_up;
		}
	}

    /* prepare the headers */
	for (Index=0;Index<NUM_HEADERS;Index++) {
		if (waveInPrepareHeader (hWave, &waveHdr[Index], sizeof(WAVEHDR)) != 0) {
			goto clean_up;
		}
    }

	/* create the progress bar */
	lpPb = BW_CalProgressBarCreate (hWnd);

	/* initialize the performance counters */
	InitPerformanceCounters ();

	/* reset to remove any buffers from the queue */
	waveInReset (hWave);

	/* add the buffers */
	for (Index=0;Index<NUM_HEADERS;Index++) {
		if (waveInAddBuffer (hWave, &waveHdr[Index], sizeof(WAVEHDR)) != 0) {
			goto clean_up;
		}
	}

	/* setup the progress bar */
	BW_CalProgressSetText (lpPb, lpText);
	BW_CalProgressBarSetPercentage (lpPb, 0);	
    
	/* start it up!!! */
	Index = 0; dwPrevious = 0;
	if (waveInStart (hWave) != 0) {
		goto clean_up;
	}

	/* init the timer */
	pcBegin();

	/* wait the specified time */
	while (TRUE) {

		/* 
		 * let other threads (like maybe the sound card driver!!!)
		 * run.
		 */
		Sleep(10);

		/* error if we take too long */
		if (pcGetTicks() > (dwCapTime*1000000) + 10000000) {
			goto clean_up;
		}

		if (pcGetTicks() - dwPrevious >= 1000000) {
		
			UINT nPercent = (UINT)(((double)(pcGetTicks()/1000000) / (double)dwCapTime) * 100.0);

			BW_CalProgressBarSetPercentage (lpPb, min (nPercent, 100));	
			dwPrevious = pcGetTicks();
		}


		/* time to add the next buffer??? */
		if (waveHdr[Index].dwFlags & WHDR_DONE)	{
				
			waveHdr[Index].dwBufferLength  = 0x10000;
			waveHdr[Index].dwBytesRecorded = 0;
			waveHdr[Index].dwFlags &= ~WHDR_DONE;
				
			if (waveInAddBuffer (hWave, &waveHdr[Index], sizeof(WAVEHDR)) != 0) {
				goto clean_up;
			}

			Index = (Index + 1) % NUM_HEADERS;
		}

		mmTime.wType = TIME_BYTES;
		if (waveInGetPosition (hWave, &mmTime, sizeof(MMTIME)) != 0) {
			goto clean_up;
		}

		if (mmTime.u.cb != cbLast) {

			cbLast = mmTime.u.cb;

			if (cbLast >= dwRecordBytes) {

				to = 0;
				while (to++ < TIMEOUT_COUNT && cbLast == mmTime.u.cb) {
					mmTime.wType = TIME_BYTES;
					waveInGetPosition (hWave, &mmTime, sizeof(MMTIME));
				}

				dwTime = pcGetTicks();
				cbLast = mmTime.u.cb;

				/* we're 100% done */
				BW_CalProgressBarSetPercentage (lpPb, 100);	
	
				dfCalc = ((double)(((double)(cbLast-cbStart)/(double)Format.nBlockAlign)/(double) dwTime)*(double)1000000.0);
				dfDiff = dfCalc/(double)(Format.nSamplesPerSec);

				bCaptureOK = TRUE;
				break;
			}

			if (cbStart == 0) {

				to = 0;
				while (to++ < TIMEOUT_COUNT && cbLast == mmTime.u.cb) {
					mmTime.wType = TIME_BYTES;
					waveInGetPosition (hWave, &mmTime, sizeof(MMTIME));
				}

				pcBegin();
				cbStart = cbLast = mmTime.u.cb;
			}
		}
	}
		
clean_up:

	/* reset to remove all buffers from the queue */
	waveInReset (hWave);

	/* unprepare and free the wave buffers */
	for (Index=0;Index<NUM_HEADERS;Index++) {
			
		if (waveHdr[Index].lpData != NULL) {		
			
			if (waveHdr[Index].dwFlags & WHDR_PREPARED)	{
				waveInUnprepareHeader (hWave, &waveHdr[Index], sizeof(WAVEHDR));                      
			}

			GlobalFreePtr(waveHdr[Index].lpData);
		}
	}
	
	/* close up the device */
    waveInClose (hWave);

	/* destroy the progress bar */
	if (lpPb != NULL) {
		BW_CalProgressBarDestroy (lpPb);
	}

	/* did we capture anything */
	if (!bCaptureOK) {
        return FALSE;
	}

	*lpdfCalFactor = dfDiff;
	return TRUE;
}


/*
 * exported audio/video synchronization calibration
 */ 
double BW_CalibrateAudio (HWND hWnd)
{
	static char *szIndex[] = {"First", "Second", "Third", "Fourth", "Fifth", "Sixth", "Seventh", "Eighth", "Ninth", "Tenth"};
	double dfDiff=0.0, dfLast = 0.0;
	BOOL bCaptureOK = FALSE, bReturn = FALSE;
	char buf[80];
	DWORD Count;

	while (TRUE) {

		/* do it until we get a stable calibration */
		for (Count=0;Count<MAX_CAL_TRIES;Count++) {

			/* set the sub caption */
			if (MAX_CAL_TRIES > 1) {
				sprintf (buf, "%s pass audio/video synchronization calibration...", szIndex[Count]);
			} else {
				sprintf (buf, "Performing audio/video synchronization calibration...");
			}

			/* save the last value */
			dfLast = dfDiff;

			/* get a new cal factor */
			if (!(bReturn = HW_CalibrateAudio (hWnd, buf, &dfDiff))) {
				bCaptureOK = FALSE;
				break;
			}

			/* 
			 * if we're suposed to try more than once, 
			 * make sure we get two values in a row
			 * that are VERY close to each other
			 */
			if (MAX_CAL_TRIES > 1) {

				if (Count > 0) {
					if (fabs (dfDiff - dfLast) <= CAL_FACTOR) {
						dfDiff = (dfDiff+dfLast)/2;
						bCaptureOK = TRUE;
						break;
					}
				}	
			} else {
				bCaptureOK = TRUE;
				break;
			}						
		}

		/* did we get a valid calibration value??? */
		if (!bCaptureOK) {
			int nRet = BW_CalErrorDialog (NULL);
			if (nRet == IDOK) 
				return 1.0;
		} else 
			break;
	}

	return (dfDiff);
}


/*
 * Hardware specific function for displaying any 
 * required driver installation dialog box.
 */
LRESULT 
DriverInstallDialog (HWND hWnd, LPCONFIG_INFO lpGeneric)
{
	/* nothing to do here... */
	return DRVCNF_OK;
}


/*
 * Hardware specific function for displaying any required driver properties 
 * dialog box. 
 */
LRESULT
DriverPropertiesDialog (HWND hWnd, LPCONFIG_INFO lpGeneric)
{
    BW_PROPERTYOPTIONS FAR *lpOptions = (BW_PROPERTYOPTIONS FAR*) lpGeneric;
    DEVID DevID = NULL;
	
    /* verify structure size */
    if (lpOptions == NULL || lpOptions->ulStructSize < sizeof(BW_PROPERTYOPTIONS)) {
		return DRVCNF_CANCEL;
    }

	if (BroadwayOpenDevice(&DevID,0) == VT_NO_ERROR) {
		BroadwayGetHardwareVersion(DevID,lpOptions->szHardwareRevision,sizeof(lpOptions->szHardwareRevision));
		BroadwayGetHardwareSerial(DevID,lpOptions->szSerialNumber,sizeof(lpOptions->szSerialNumber));

		if (!lstrcmpi ("Unknown", lpOptions->szSerialNumber)) {
			wsprintf (lpOptions->szSerialNumber, "See user guide");
		}
		
		BroadwayCloseDevice(DevID);
	} else {
		wsprintf (lpOptions->szHardwareRevision, "No board found");
		wsprintf (lpOptions->szSerialNumber, "Unknown");
		memset (lpOptions->szSerialNumber, 0, sizeof(lpOptions->szSerialNumber)); 
	}

	return BW_CapturePropertiesDialog (hWnd, lpOptions);
}



/* called for any device-specific clean up
 * we need to disable overlays
 */
BOOL
HW_Cleanup(VCUSER_HANDLE vh)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);

    /* verify harware info */
    if (!IsBadHWINFO(lpHw)) {
		
        /* clean up the hardware */
        if (lpHw->DevID != 0) {
            
            /* make sure capture and overlay are disabled */
            /* reset the hardware */
			lpHw->bNoRestart = TRUE;
            HW_Capture (vh, FALSE);            
			
			/* make sure the hardware is disabled */
			VT_Reset(lpHw->DevID);
			VT_SetMode(lpHw->DevID, VT_IDLE);

            /* delete the device */
            VT_DeleteDevice(lpHw->DevID);
        }

		/* free the single frame copy buffer */
		if (lpHw->lpLastFrame != NULL) {
			GlobalFreePtr (lpHw->lpLastFrame);
		}

        /* release the single frame capture event */
		if (lpHw->hFrameEvent != NULL) {
            CloseHandle(lpHw->hFrameEvent);             
        }

        /* delete the overlay engine */
        if (lpHw->voh != 0) {
		    VO_StopOverlay (lpHw->voh);
            VO_TerminateOverlay (lpHw->voh);
            lpHw->voh = 0;
        }

		/* delete the realtime loop critical section */
		if (lpHw->bCriticalInit) {
			DeleteCriticalSection (&lpHw->RTLoopCritical);
		}

        /* zero out the structure */
        FillMemory (lpHw, sizeof(HWINFO), 0);
		glpHwCurrent = NULL;
    }

	return(TRUE);
}

/*
 * called on the first device-open. We use this function to do
 * video mode-specific inits, since VC_GetVideoMode cannot be
 * called at DriverEntry time (see vckernel.h).
 *
 * If the video mode is one we don't support, then we will disable overlay
 * support by setting the relevant callbacks to NULL. Thus we can
 * still do capture even if we can't do overlay.
 *
 */
BOOL
HW_DeviceOpen(VCUSER_HANDLE vh)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);

    /* verify the hardware info */
    if (IsBadHWINFO(lpHw)) {
        return FALSE;        
    }
 
    return (VO_InitializeOverlay (lpHw, &lpHw->voh));
}                    


/*
 * initialize the hardware to a known state.
 */
BOOL
HW_Init(VCUSER_HANDLE vh, LPTSTR lpDriverName, DWORD DeviceIndex)
{
    BW_CAPTURE_CONFIG Config;
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	BOARD_STATE HWState;
	DWORD start, dwCal = 0, dwInstall = 0;
    DWORD Type, Length, lRet;
    BOOL bResult = TRUE;
	HKEY hkOptions;
	double dfCal;
    
	/* verify the hardware info */
    if (IsBadHWINFO(lpHw)) {
        return FALSE;        
    }

	/* init the realtime loop critical section */
	InitializeCriticalSection (&lpHw->RTLoopCritical);
	lpHw->bCriticalInit = TRUE;
    
	/* initialize the video tool library for this device */
    if (VT_OpenDevice(lpDriverName, DeviceIndex, &lpHw->DevID) != VT_NO_ERROR) {
        lpHw->DevID = 0;
        return FALSE;
    }

#ifdef _DEBUG
    /* reset the hardware */
    if (VT_Reset(lpHw->DevID) != VT_NO_ERROR) {
        return FALSE;
    }
#else
	/* 
	 * before we go any further, verify we can take over ownership
	 * of the board cleanly.  If that trickey little codec is running,
	 * give it a few tries, we'll get it eventually.  if the recoder
	 * of another instance of the capture driver is running, we'll
	 * fail, as we should!!!
	 */
	start=timeGetTime();
	while (TRUE) {

		if (Timeout(start,100)) {
	        VT_DeleteDevice(lpHw->DevID);
		    lpHw->DevID = 0;
	        return FALSE;
		}

		if (VT_GetHardwareState(lpHw->DevID, &HWState) != VT_NO_ERROR) {
	        VT_DeleteDevice(lpHw->DevID);
		    lpHw->DevID = 0;
	        return FALSE;
		} 
		
		if (HWState == INITIALIZED || HWState == CODEC_INITIALIZED) {
			break;
		} else {
			Sleep(1);
		}						     
	}

#endif

    /* reset the hardware */
    if (VT_Reset(lpHw->DevID) != VT_NO_ERROR) {
        return FALSE;
    }

	/* don't let anyone else touch our hardware!!! */
	//VT_AquireMutex();

    /* fill in the board register and capture buffer information */ 
    if (VT_GetDeviceMemory(lpHw->DevID, &(LPVOID)lpHw->lpRegs, NULL) != VT_NO_ERROR) { 
		bResult = FALSE;
		goto _release_mutex_;
    }

    if (VT_GetHardwareBuffer(lpHw->DevID, &lpHw->dwPhysicalAddr, &lpHw->lpLinearAddr, NULL, &lpHw->dwMemorySize)) {
		bResult = FALSE;
		goto _release_mutex_;
    }
	
    /* make sure we have enough buffer space */
    if (lpHw->dwMemorySize < (REQUIRED_BUFFER_SIZE)) {
		bResult = FALSE;
		goto _release_mutex_;
    }

	/* get the hardware revision */
	VT_GetRevisionID(lpHw->DevID, &lpHw->RevisionID);
    
	/* Initialize the hardware structure */
	lpHw->DeviceIndex = DeviceIndex;
	lpHw->vh          = vh;
	lpHw->IsBroadway  = TRUE; 
	lpHw->bCapture    = FALSE;

	/* initize the options structures */
	lpHw->CaptureOptions.ulStructSize	  = sizeof(BW_CAPTUREOPTIONS);
	lpHw->CapturePreferences.ulStructSize = sizeof(BW_CAPTURE_PREFERENCES);
	lpHw->CaptureOptions.lpPreferences	  = &lpHw->CapturePreferences;

	lpHw->RecodeOptions.ulStructSize	  = sizeof(BW_RECODEOPTIONS);
	lpHw->RecodePreferences.ulStructSize  = sizeof(BW_RECODE_PREFERENCES);
	lpHw->RecodeOptions.lpPreferences	  = &lpHw->RecodePreferences;

    /* Initialize the capture info structure */
    lpHw->CaptureInfo.StructSize  = sizeof(CAPTUREINFO);
    lpHw->CaptureInfo.Quality     = 10000;
    lpHw->CaptureInfo.Padding     = FALSE;
    lpHw->CaptureInfo.PassMode    = 2;
    lpHw->CaptureInfo.SourceIsVcr = TRUE;
    lpHw->CaptureInfo.Aspect      = 12;
    lpHw->CaptureInfo.ShiftLeft   = 8;
    lpHw->CaptureInfo.Blanklines  = 0;
    lpHw->CaptureInfo.FilterValue = 0;
    lpHw->CaptureInfo.ShiftUp	  = -8;

	/* initialize the video buffers */
	if (!HW_InitializeBuffers(vh)) {
		bResult = FALSE;
		goto _release_mutex_;
	}

    /* allocate one capture buffer for single frame copies */
	lpHw->dwLastFrameMax = MAX_CAPT_SIZE*sizeof(DWORD);
    lpHw->lpLastFrame = GlobalAllocPtr (GPTR, lpHw->dwLastFrameMax);
    if (lpHw->lpLastFrame == NULL) {
		bResult = FALSE;
		goto _release_mutex_;
	}

    /* create an event for single frame capture */
    lpHw->hFrameEvent = CreateEvent (NULL, FALSE, FALSE, NULL);
    if (lpHw->hFrameEvent == NULL) {
		bResult = FALSE;
		goto _release_mutex_;
    }

    /* open the registry profile */
    Config.lpProf = VC_OpenProfileAccess("vtool", DeviceIndex);
    if (Config.lpProf == NULL) {
		bResult = FALSE;
		goto _release_mutex_;
    }
    
    /* get the log file setting from the registry */        
    lpHw->bLogFileEnable = VC_ReadProfileUser(Config.lpProf, "Settings", "EnableLogFile", FALSE);
    VC_WriteProfileUser (Config.lpProf, "Settings", "EnableLogFile", lpHw->bLogFileEnable);
             
    /* configure the default settings */
	Config.ulStructSize = sizeof(BW_CAPTURE_CONFIG);
	Config.nMode		= MODE_FETCH_SETTINGS;
    Config.nIsBroadway	= TRUE;
	Config.nWndParent   = (long)NULL;

    if (!HW_ConfigFormat(vh, (LPCONFIG_INFO)&Config)) {
	    VC_CloseProfileAccess(Config.lpProf);
	}

    /* close the profile */
    VC_CloseProfileAccess(Config.lpProf);

  	/* load first Master1 DMA buffer (passthru/overlay) */
   	VT_LoadDma(lpHw->DevID
   	           ,MASTER1
   	           ,lpHw->PassBufs.Bufs[0].dwLength
   	           ,lpHw->PassBufs.Bufs[0].dwPhys
   	           );

   	/* load first Master2 DMA buffer (I-frame capture) */
   	VT_LoadDma(lpHw->DevID
   	           ,MASTER2
   	           ,lpHw->CaptBufs.Bufs[0].dwLength
   	           ,lpHw->CaptBufs.Bufs[0].dwPhys
   	           );

	_release_mutex_:
	
	//VT_ReleaseMutex();

	/* 
	 * this is stuff we only want to waste time doing if we
	 * survived to this point!!!
	 */
	if (bResult) {    
    
		/* See if we can get a registry handle to our settings data */
		lRet = RegOpenKey(HKEY_LOCAL_MACHINE, STR_BROADWAY_HARDWARE_SUBKEY, &hkOptions);
		if (lRet  == ERROR_SUCCESS) {

			/* try to get the current audio cal setting */
			Length = sizeof(dwCal);  
			lRet = RegQueryValueEx(hkOptions
				                   ,STR_AUDIO_CAL_FACTOR
					               ,NULL
						           ,&Type
							       ,(PUCHAR)&dwCal
								   ,&Length
	                               );

		    if (lRet != ERROR_SUCCESS || Type != REG_DWORD) {
				dwCal = 0;
			}

			/* try to get the current driver install setting */
			Length = sizeof(dwInstall);  
	        lRet = RegQueryValueEx(hkOptions
		                           ,STR_DRIVER_INSTALL
			                       ,NULL
				                   ,&Type
					               ,(PUCHAR)&dwInstall
						           ,&Length
							       );

			RegCloseKey(hkOptions);

			if (lRet  != ERROR_SUCCESS || Type != REG_DWORD) {
				dwInstall = 0;
			}
		}

		/* if this is first time after install, do audio calibration */
		if (dwCal == 0 || dwInstall == 0) {

			dfCal = BW_CalibrateAudio (NULL);
			dwCal = (DWORD)(dfCal * (double)1000000000.0);

			/* write the new value to the registry */
		    lRet = RegCreateKey(HKEY_LOCAL_MACHINE, STR_BROADWAY_HARDWARE_SUBKEY, &hkOptions);
			if (lRet == ERROR_SUCCESS) {

				lRet = RegSetValueEx(
						    hkOptions,
							STR_AUDIO_CAL_FACTOR,
							0,
							REG_DWORD,
							(const unsigned char *)&dwCal,
							sizeof(dwCal));

				dwInstall = 1L;
				lRet = RegSetValueEx(
					    hkOptions,
						STR_DRIVER_INSTALL,
						0,
						REG_DWORD,
						(const unsigned char *)&dwInstall,
						sizeof(dwInstall));

				RegCloseKey(hkOptions);
			}
		}

		lpHw->dfAudioFudgeFactor = (double)dwCal/(double)1000000000.0;

		/* get the handle to the vidcap32 dll */
		lpHw->hVfw = GetModuleHandle ("avicap32.dll");

		/* initialize the performance counters */
		InitPerformanceCounters ();

		/* set the global hardware pointer for this process */
		glpHwCurrent = lpHw;
	}
			   

    /* all set up ok */
    return bResult;
}


BOOL 
HW_InitializeBuffers (VCUSER_HANDLE vh)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
    DWORD dwPhysicalAddr, dwMemorySize, i;
    DWORD FAR *lpLinearAddr;

    /* verify the hardware info */
    if (IsBadHWINFO(lpHw)) {
        return FALSE;        
    }

	/* save the base addresses */
	dwPhysicalAddr = lpHw->dwPhysicalAddr;	
	lpLinearAddr = lpHw->lpLinearAddr;		

	/* fill in the capture buffer information */
	dwMemorySize = lpHw->dwMemorySize;// - 1;
	lpHw->dwCaptHead = lpHw->dwCaptTail = 0;

	lpHw->CaptBufs.dwNumberBufs = (lpHw->bOverlay) ?NUM_CAPT_BUFS:1;
	lpHw->CaptBufs.dwCurrentBuf = 0;
    lpHw->CaptBufs.dwDmaIndex   = 0; 

    for (i=0;i<lpHw->CaptBufs.dwNumberBufs;i++) {
        lpHw->CaptBufs.Bufs[i].lpData		= lpLinearAddr;
        lpHw->CaptBufs.Bufs[i].dwStartIndex = 0;
        lpHw->CaptBufs.Bufs[i].dwPhys		= dwPhysicalAddr;
        lpHw->CaptBufs.Bufs[i].dwLength		= dwMemorySize;
        lpHw->CaptBufs.Bufs[i].dwFilled		= 0;
		
        lpLinearAddr += MAX_CAPT_SIZE;
        dwPhysicalAddr += MAX_CAPT_SIZE * sizeof(DWORD);
    }

    /* 
	 * fill in the passthru buffer information - we should include a 
	 * passthru buffer list even if not doing overlay - so we can program
	 * MASTER 1 with a valid address.
     */
	dwPhysicalAddr = lpHw->dwPhysicalAddr+((MAX_CAPT_SIZE*NUM_CAPT_BUFS) * sizeof(DWORD));		
	lpLinearAddr = lpHw->lpLinearAddr+(MAX_CAPT_SIZE*NUM_CAPT_BUFS);		

    lpHw->PassBufs.dwNumberBufs = NUM_PASS_BUFS;
	lpHw->PassBufs.dwCurrentBuf = 0;
    lpHw->PassBufs.dwDmaIndex   = 0; 
    
    for (i=0;i<lpHw->PassBufs.dwNumberBufs;i++) {
        lpHw->PassBufs.Bufs[i].lpData		= lpLinearAddr;
        lpHw->PassBufs.Bufs[i].dwStartIndex = 0;
        lpHw->PassBufs.Bufs[i].dwPhys		= dwPhysicalAddr;
        lpHw->PassBufs.Bufs[i].dwLength		= MAX_PASS_SIZE;
        lpHw->PassBufs.Bufs[i].dwFilled		= 0;

        lpLinearAddr += MAX_PASS_SIZE;
        dwPhysicalAddr += MAX_PASS_SIZE * sizeof(DWORD);
    }
	
	return TRUE;
}



