/*
 * interrupt.c
 *
 *
 * 32-bit Video Tool driver
 * timer interrupt service routines 
 * for Data Translation Video Tool.
 *
 * Mark Vrabel, October 95
 */

#include <vcdriver.h>
#include <bwuser.h>
#include "hardware.h"													
#include <mmtimers.h>

#define MAX_LOSTFRAMES			(1)
#define FIFO_RESETCOUNT			(0)
#define SKIPFRAME				(3)

#define NO_RESTART_NOFRAME_COUNT	(3)
#define IN_RESTART_NOFRAME_COUNT	(10)
#define MAX_NOFRAME_TIME            (8000)


/* define local functions */
ULONG __inline HW_TransferFrame(LPHWINFO lpHw, LPUCHAR lpOutput, ULONG OutputLength, LPVIDEO_BUFFER lpvb, BOOL FAR*lpKeyFame);
BOOL HW_SkipFrame(LPHWINFO lpHw, LPVIDEO_BUFFER lpvb, DWORD dwSize);


__inline DWORD TruncateingMulDiv (DWORD dwNumber, DWORD dwNumerator, DWORD dwDenominator) 
{
	DWORDLONG Result = UInt32x32To64(dwNumber, dwNumerator);
	return (DWORD)(Result / (DWORDLONG)dwDenominator);
}

int dead = 0;

BOOL HW_IsDead(LPHWINFO lpHw, DWORD dwSystemTime)
{
	DWORD Dst, dwTimeSpan, dwCount, dwMaxCount;
	DWORD dwCurrentTime = dwSystemTime - lpHw->dwTimeOffset;

   /* 
	* usefull when debuging the restart code. change #if 0 to #if 1 to 
	* force periodic restarts.
	*/
#if 0
	if (++dead % 2000 == 0) {
		lpHw->BadToggle++;
		return TRUE;
	}

	return FALSE; 

#endif

	/* 
	 * subtract any offset and calculate time since last frame 
	 */
	dwTimeSpan = (dwCurrentTime-lpHw->dwTimePrevious);
	
	/* if we haven't got anything for five seconds, we are really, really dead */
	if (dwTimeSpan > MAX_NOFRAME_TIME) {

		lpHw->dwCaptureStopped++;
		return FALSE;
	} 

	/* 
	 * if we're looking at less than an interrupt time here or we haven't gotten
	 * an interrupt yet - or we in the middle of restarting, we're not dead!!!
	 */
	if (dwTimeSpan * 1000 < lpHw->dwMicroPerInterrupt || lpHw->dwInterruptCount == 0) 
	{
		return FALSE;
	}

	/* 
	 * try to read c-cube dram - if we can't we're dead for sure. this call 
	 * generates an exception on failure, so we need to catch the exception
	 * to fail gracefully.
	 */
	__try 
	{
	VT_Read(lpHw->DevID, VR_DRAM, &Dst, 0x39f, 1);
	}
	
	__except ( _exception_code()==VT_ERR_IREG_TIMEOUT)
	{
		lpHw->BadToggle++;
		return TRUE;
	}


	if (lpHw->bRestart) {
		if (dwSystemTime > lpHw->dwTimeRestart) 
			dwTimeSpan = dwSystemTime - lpHw->dwTimeRestart;
		else
			dwTimeSpan = 0;
			
		dwCount = TruncateingMulDiv (dwTimeSpan, 1000L, lpHw->dwMicroPerInterrupt);
		dwMaxCount = IN_RESTART_NOFRAME_COUNT;
	} else {
		dwCount = TruncateingMulDiv (dwTimeSpan, 1000L, lpHw->dwMicroPerInterrupt);
		dwMaxCount = NO_RESTART_NOFRAME_COUNT;
	}

	/* 
	 * if we got this far, concider us dead if we haven't gotten
	 * any data in a while - assumue this function only gets called
	 * when data count is 0.
	 */
	if (dwCount > dwMaxCount) {
		lpHw->BadToggle++;
		return TRUE;
	}

    return FALSE;
}


void
HW_DynamicFifoReset(LPHWINFO lpHw)
{
	DWORD dwEnd;

	if (lpHw->dwLostFrames == 0) 
		lpHw->dwSkipFrame = SKIPFRAME;
	else 
		lpHw->dwSkipFrame = 0;

	if (++lpHw->dwLostFrames < MAX_LOSTFRAMES) 
		return;
	
	dwEnd = timeGetTime();
	dwEnd += 20;

    lpHw->lpRegs->Status = VSYNC_FLAG;
	while (!lpHw->lpRegs->Status & VSYNC_FLAG) {
		if (timeGetTime() >= dwEnd)
			break;
	}

	VT_SetupFifo(lpHw->DevID, 63, 8);
	VT_ReadModifyWrite(lpHw->DevID, CR_REGS, CTRL2, 0xff, 0x10);
	lpHw->lpRegs->Status = DONE_FLAG_1+DONE_FLAG_2;
					
	lpHw->ResetFifoCount = FIFO_RESETCOUNT;
	lpHw->dwSkipFrame = SKIPFRAME;
	lpHw->dwLostFrames = 0L;
	lpHw->FifoReset++;
}


BOOL
HW_CheckPreviewOverRun (LPHWINFO lpHw, LPVIDEO_BUFFER lpvb) 
{
	DWORD dwLast, dwCurrent, Index;

	/* check to see if we over ran the buffer */
	if (lpvb->lpData[lpHw->dwLast] != 0xffffffff) {

		if (lpHw->bVideoIn) {
			lpHw->CopyOverRun++;        
		}

		dwCurrent = lpHw->lpRegs->MasterStart2;
		dwCurrent = (dwCurrent - lpHw->dwPhysicalAddr)/sizeof(DWORD);

		/* 
	     * wait for the transfer to complete so 
	     * we can start fresh the next time through
	     */
	    do {
			dwLast = dwCurrent;
		 		
			for (Index=0;Index<100;Index++) {
				dwCurrent = lpHw->lpRegs->MasterStart2;
			}
		
			dwCurrent = lpHw->lpRegs->MasterStart2;
			dwCurrent = (dwCurrent - lpHw->dwPhysicalAddr)/sizeof(DWORD);
        
		} while (dwLast != dwCurrent);
			
		/* on copy overrun, call it a wash and reset the tail */
		lpHw->dwCaptTail = dwCurrent;
		lpHw->dwLast = (lpHw->dwCaptTail + (lpvb->dwLength - 1)) % lpvb->dwLength;
		lpvb->lpData[lpHw->dwLast] = 0xffffffff;
		return TRUE;
	}

	return FALSE;
}


/* 
 * service the timer event for no-passthru mode. We are captureing into
 * circular buffers and must account for buffer rollover and mutiple
 * frames in the buffer.
 */
DWORD  
HW_InterruptNoPassthru(VCUSER_HANDLE vh, LPHWINFO lpHw, DWORD Status)
{
    LPVIDEO_BUFFER lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf];
	DWORD *lpSearch = lpvb->lpData;
    DWORD dwCurrent, dwLast, dwSize, Index, dwFrameSize, dwStart;
	BOOL bFoundFirstPicture = FALSE, bFoundNextPicture = FALSE;
    DWORD dwIncrement = 0x01;
	DWORD dwHeaderPictureStart = 0xffffffff;
	BOOL bDiscard = FALSE;

	/* make sure we have at least one completed transfer */
    dwStart = lpHw->lpRegs->MasterStart2;
    dwCurrent = (dwStart - lpHw->dwPhysicalAddr)/sizeof(DWORD);

    if (dwCurrent == lpHw->dwCaptTail) {
		return 0;
	}

	if (lpHw->bRestart) {
		OutputDebugString ("What the #$#@!!!\n");
	}

	/* 
     * wait for the transfer to complete so 
     * we can start fresh the next time through
     */
    do {
	    dwLast = dwCurrent;
		 		
		for (Index=0;Index<100;Index++) {
			dwCurrent = lpHw->lpRegs->MasterStart2;
		}
		
		dwStart = lpHw->lpRegs->MasterStart2;
		dwCurrent = (dwStart - lpHw->dwPhysicalAddr)/sizeof(DWORD);
        
	} while (dwLast != dwCurrent);

	/* 
	 * this is a bad situation that should never happen, but I have seen it
	 * occur on a Dell Optiplex dual processor machine where the broadway
	 * was not in on of the "faster" slots
	 */
	if (dwCurrent > lpvb->dwLength) {
		return 0;
	}	

	/* check for fifo overflow */
	if (Status & FIFO_OVERFLOW) {

		/* clear the fifo overflow bit */
		lpHw->lpRegs->Status = FIFO_OVERFLOW;
		if (lpHw->bVideoIn) {
			lpHw->FifoOverFlow++;        

			lpHw->bRestart = TRUE;
			lpHw->bGotIFrame = TRUE;
			RT_FlushToLastGop (lpHw);
		}

		/* on fifo overflow, call it a wash and reset the tail */
		lpHw->dwCaptTail = dwCurrent;
		lpHw->dwLast = (lpHw->dwCaptTail + (lpvb->dwLength - 1)) % lpvb->dwLength;
		lpvb->lpData[lpHw->dwLast] = 0xffffffff;

		return 0;
	} 

	/* check to see if we over ran the buffer */
	if (HW_CheckPreviewOverRun(lpHw, lpvb)) {
		return 0;
	}

	/* calculate the count since the last transfer */
	if (dwCurrent >= lpHw->dwCaptTail) {
		dwSize = dwCurrent - lpHw->dwCaptTail;
	} else {
		dwSize = (lpvb->dwLength - lpHw->dwCaptTail) + dwCurrent;
	}

	dwFrameSize = dwSize;
	Index = lpHw->dwCaptTail;
	lpHw->bMultiFrame = FALSE;

	while (dwFrameSize != 0) {
		
		if (!lpHw->bVideoIn || lpHw->CaptureMode != CAPMODE_RTMPEG) {

			switch (lpSearch[Index]) {
			
				case 0xb3010000:
					if (bFoundFirstPicture) {
						/* found a second picture in the buffer */
						bFoundNextPicture = TRUE;	
						
						/* this will give us the correct size of the first picture */
						dwCurrent = Index;
						break;
					} else {
						/* 
						 * we found the first picturein the buffer, now we can safely search
						 * on 16 dwordboundries for the next picture
						 */
						bFoundFirstPicture = TRUE;	
						dwIncrement = 0x10;

						/* this will remove the leading zero padding */
						lpHw->dwCaptTail = Index;
						break;
					}

				default:
					break;
			}
		} else {
			
			switch (lpSearch[Index]) {
			
				case 0xb3010000:

					/* 
					 * are we restarting afer an error? if we have just restarted the 
					 * c-cube, we want to skip the very first gop due to quality/line
					 * blanking problems. otherwise we are make sure we start on 
					 * the next gop.
					 */
					if (lpHw->bRestart) {
						if (!lpHw->bGotIFrame) {
							lpHw->dwTimeRestart = (pcGetTime());
							lpHw->bGotIFrame = TRUE;
						} else {
						    DWORD tcIndex = Index;
							DWORD tcSize = dwFrameSize;
							MPEG_TIME_CODE tc;
							
							/* find the gop start code... */
							while (tcSize != 0) {       
            
								if (lpSearch[tcIndex] == 0xb8010000) {
									break;
								}
        
								tcIndex = (tcIndex + 1) % lpvb->dwLength;;
								tcSize--;
							}

							/* at this point tcSize should never be 0!!! */
							if (tcSize != 0) {
								/* if skipping first gop - set the broken link flag */
								tcIndex = (tcIndex + 1) % lpvb->dwLength;;
								tc.dword = ByteSwap (lpSearch[tcIndex]);

								tc.fld.bl = 1;
							
								lpSearch[tcIndex] = ByteSwap (tc.dword);
								lpHw->bGotIFrame = FALSE;
							}
						}
					}

					if (bFoundFirstPicture) {
						/* found a second picture in the buffer */
						bFoundNextPicture = TRUE;

						/* this will give us the correct size of the first picture */
						dwCurrent = Index;
						break;
					} else {

						DWORD tmpSize = dwFrameSize;
						DWORD tmpIndex = Index;
						
						/* find the picture start code that goes with this frame */
						while (tmpSize != 0) {       
            
							if (lpSearch[tmpIndex] == 0x00010000) {
								dwHeaderPictureStart = tmpIndex;
								break;
							}
        
					        tmpSize -= min (dwIncrement, tmpSize);
							tmpIndex = (tmpIndex + min (dwIncrement, tmpSize)) % lpvb->dwLength;
						}

						/* 
						 * we found the first picture in the buffer, now we can safely search
						 * on 16 dword boundries for the next picture
						 */
						bFoundFirstPicture = TRUE;	
						dwIncrement = 0x10;
						break;
					}
				
				case 0x00010000:

					if (lpHw->bRestart && lpHw->bGotIFrame) {
						lpHw->dwTimeRestart = (pcGetTime());
						bDiscard  = TRUE;
					}

					if (bFoundFirstPicture && Index != dwHeaderPictureStart) {
						/* found a second picture in the buffer */
						bFoundNextPicture = TRUE;

						/* this will give us the correct size of the first picture */
						dwCurrent = Index;
						break;
					} else {
						/* 
						 * we found the first picture in the buffer, now we can safely search
						 * on 16 dword boundries for the next picture
						 */
						bFoundFirstPicture = TRUE;	
						dwIncrement = 0x10;
						break;
					}

				default:
					break;
			}
		}	
	
		/* 
		 * if we found a second frame - we're done.  we'll be back immediately 
		 * to get the next frame.
		 */
		if (bFoundNextPicture) {
			lpHw->bMultiFrame = TRUE;
			if (lpHw->bVideoIn) {
				lpHw->MultiFrame++;
			}
			break;
		}

        dwFrameSize -= min (dwIncrement, dwFrameSize);
	    Index = (Index + min (dwIncrement, dwFrameSize)) % lpvb->dwLength;
	}

	/* mark the dword before the old tail so we can check for buffer overrun */
	lpHw->dwLast = (lpHw->dwCaptTail + (lpvb->dwLength - 1)) % lpvb->dwLength;
	lpvb->lpData[lpHw->dwLast] = 0xffffffff;

	/* mark the buffer's start index */
	lpvb->dwStartIndex = lpHw->dwCaptTail;

	/* calculate the the new size */
	if (dwCurrent >= lpHw->dwCaptTail) {
		dwSize = dwCurrent - lpHw->dwCaptTail;
	} else {
		dwSize = (lpvb->dwLength - lpHw->dwCaptTail) + dwCurrent;
	}

	/* set the new tail */
	lpHw->dwCaptTail = (lpHw->dwCaptTail + dwSize) % lpvb->dwLength;

	/* if no picture found, we're hosed!!! */
	if (!bFoundFirstPicture) {
		if (lpHw->bVideoIn) {
			lpHw->NoStartCode++;
		}
		return 0;
	}

	if ((lpHw->bRestart && lpHw->bGotIFrame) || bDiscard) {
		return 0;
	}

	return dwSize;
}


/* 
 * service the timer event for passthru mode. We are captureing into
 * multiple buffers for master 1 (passthru) and a single buffer for
 * master 2 (I-frame capture).
 */
DWORD 
HW_InterruptPassthru(VCUSER_HANDLE vh, LPHWINFO lpHw, DWORD Status)
{
    LPVIDEO_BUFFER lpvb;
    DWORD dwSize = 0;

    /* if no dma done's - get out quick!!! */
    if (!(Status & (DONE_FLAG_1+DONE_FLAG_2))) {
        return 0;
    }

    /* clear any done bits in the status register */
    lpHw->lpRegs->Status = DONE_FLAG_1+DONE_FLAG_2;

    /* 
     * Passthru Done:
     *
     * check Master1 DMA (passthru/overlay) 
     */
    if (Status & DONE_FLAG_1) {
		/* 
		 * NOTE: Handleing of the Master DMA 1 interrupt has
		 * been moved to overlay.cpp for better screen update
		 * control.
		 */
		if (lpHw->bVideoIn) {
			lpHw->bCanDrawNow = FALSE;
		} else {
			lpHw->bCanDrawNow = TRUE;
		}
    } 
    
    /* 
     * Buffer Done:
     *
     * check Master2 DMA (I-frame capture) 
     */
    if (Status & DONE_FLAG_2) {

		DWORD dwStart, dwPhys;

		/* load Master2 DMA (I-frame capture) */
		if (lpHw->CaptBufs.dwNumberBufs > 1) 
		{
			lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwDmaIndex];

			VT_LoadDma(lpHw->DevID
		               ,MASTER2
			           ,lpvb->dwLength
				       ,lpvb->dwPhys
					   );
		}
		
		lpHw->CaptBufs.dwDmaIndex = (lpHw->CaptBufs.dwDmaIndex + 1) % lpHw->CaptBufs.dwNumberBufs;

        /* set the current buffer */
        lpHw->CaptBufs.dwCurrentBuf = (lpHw->CaptBufs.dwCurrentBuf + 1) % lpHw->CaptBufs.dwNumberBufs;
        lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf];
		
        /* check for errors */
		dwStart = lpHw->lpRegs->MasterStart2;
		dwPhys = lpvb->dwPhys;

		if (dwStart > dwPhys && dwStart < dwPhys + MAX_CAPT_SIZE*sizeof(DWORD))	{
			lpHw->lpRegs->Status = DONE_FLAG_1+DONE_FLAG_2;
			return 0;
		}

		if (Status & FIFO_OVERFLOW) {
            lpHw->lpRegs->Status = Status & FIFO_OVERFLOW;
            lpHw->FifoOverFlow++;
            return 0;
        } 

		/* skip the reset count after fifo reset */
		if (lpHw->ResetFifoCount != 0) {
			lpHw->ResetFifoCount--;
			return 0;
		}

		/* skip the reset count after fifo reset */
		if (lpHw->dwSkipFrame != 0) {
			lpHw->dwSkipFrame--;
			if (lpHw->dwSkipFrame == 0)
				return 0;
		}

		/* get the buffer size */
        dwSize = lpHw->lpRegs->Status & 0x3ffff;
        dwSize = lpvb->dwLength - dwSize;
        
        /* store the frame info */
        lpvb->dwStartIndex = 0;

        /* 
         * scedule an event to try and draw a passthru frame when not in the 
         * middle of a data transfer!!! 
         */
		lpHw->bCanDrawNow = TRUE;
		if (!(lpHw->CaptureMode == CAPMODE_RTMPEG && lpHw->bVideoIn)) {
			VO_DrawFrame(lpHw->voh);
		}
    }

    return dwSize;
}


/*
 * repeat a stream capture frame for audio/video synchronization.
 * this is done so the repeated frame does not appear as a dropped 
 * frame to the user
 */ 
ULONG 
HW_RepeatFrame(
  VCUSER_HANDLE vh,
  LPUCHAR lpInput,
  ULONG InputLength,
  LPUCHAR lpOutput,
  ULONG OutputLength,
  LPULONG lpTimeStamp
)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	DWORD dwX, FrameType, dwTemporalReference;
	DWORD dwIndex = 0;

	static dwTemporalOffset;

    /* verify hardware info */
    if (IsBadHWINFO(lpHw)) {
        return 0;
    }

	/* 
	 * for real-time capture we can only repeat B-frames, so we have to
	 * find the GOP start, find a B-frame to repeat, and fixup the 
	 * temporal references.
	 */
	if (lpHw->CaptureMode == CAPMODE_RTMPEG) {
		DWORD *lpSearch = (DWORD*)lpInput;
		DWORD dwSize = InputLength;

		while (dwSize != 0) {       
            
			if (lpSearch[dwIndex] == 0x00010000) {
				break;
			}
        
			dwIndex += 1;
			dwSize--;
		}
				
		if (dwSize == 0) {
			return 0;	
		}

		dwIndex += 1;
		dwX = ByteSwap (lpSearch[dwIndex]);
		FrameType = (dwX & 0x00380000) >> 19;
		dwTemporalReference  = (dwX & 0xFFC00000) >> 22;

		lpHw->bRepeatFrame = TRUE;
		if (lpHw->bGotIFrame) {
		
			switch (dwTemporalReference) {                                                            
			                                                                                              
				case 0:
					if (lpHw->dwRepeatFrames == 0) {
						return 0;
					} 
					
					lpHw->dwRepeatFrames--;
					dwTemporalReference += ++dwTemporalOffset;
					break;

				case 1:
				case 3:
				case 4:
				case 5:
				case 6:
				case 8:
					dwTemporalReference += dwTemporalOffset;
					dwX &= ~0xFFC00000;
					dwX |= (dwTemporalReference << 22);
					lpSearch[dwIndex] = ByteSwap (dwX);
					return 0;

				case 7:
					dwTemporalReference += dwTemporalOffset;
					dwX &= ~0xFFC00000;
					dwX |= (dwTemporalReference << 22);
					lpSearch[dwIndex] = ByteSwap (dwX);

					lpHw->bGotIFrame = FALSE;
					lpHw->bRepeatFrame = FALSE;
					return 0;

				default:
					return 0;
			}

		} else if (FrameType == I_FRAME) {

			/* the I-frame has a temporal reference of 2 */
			dwTemporalReference += lpHw->dwRepeatFrames;
			dwTemporalOffset = 0;

			dwX &= ~0xFFC00000;
			dwX |= (dwTemporalReference << 22);
			lpSearch[dwIndex] = ByteSwap (dwX);
			lpHw->bGotIFrame = TRUE;
			return 0;
		} else {
			return 0;
		}
	} else {
		
		/* I-frame capture always repeats if count is not 0 */
		if (lpHw->dwRepeatFrames == 0) {
			return 0;
		} else {
			lpHw->dwRepeatFrames--;
		}
	}
		
	/* update the video time to the next frame */
	lpHw->dwInterruptCount += 1;
	lpHw->dwVideoTime = MulDiv (lpHw->dwInterruptCount, lpHw->dwMicroPerInterrupt, 1000L);

	/* set the time stamp */
	*lpTimeStamp = lpHw->dwVideoTime;

	/* verify buffer length */
	if (OutputLength < InputLength) {
		return 0;
	}
	if (InputLength == 0) {
		return 0;
	}
	
	/* copy the buffer */ 
	CopyMemory (lpOutput, lpInput, InputLength);
	lpHw->Repeated++;                                                                                      

	if (lpHw->CaptureMode == CAPMODE_RTMPEG) {
		dwX &= ~0xFFC00000;
		dwX |= (dwTemporalReference << 22);
		((DWORD*)lpOutput)[dwIndex] = ByteSwap (dwX);
	}
	
	/* return the count */
	return InputLength;
}


BOOL
HW_SkipFrame(LPHWINFO lpHw, LPVIDEO_BUFFER lpvb, DWORD dwSize)
{
	/* 
	 * for real-time capture we can only skip B-frames, so we have to
	 * find the GOP start, find a B-frame to skip, and fixup the 
	 * temporal references.
	 */
	if (lpHw->CaptureMode == CAPMODE_RTMPEG) {

		DWORD dwX, FrameType, dwTemporalReference;
		DWORD dwIndex = lpvb->dwStartIndex;

		while (dwSize != 0) {       
            
			if (lpvb->lpData[dwIndex] == 0x00010000) {
				break;
			}
        
			dwIndex = (dwIndex + 1) % lpvb->dwLength;
			dwSize--;
		}

		if (dwSize == 0) {
			return FALSE;
		}
									   
		dwIndex = (dwIndex + 1) % lpvb->dwLength;
		dwX = ByteSwap (lpvb->lpData[dwIndex]);
		FrameType = (dwX & 0x00380000) >> 19;
		dwTemporalReference  = (dwX & 0xFFC00000) >> 22;

		lpHw->bSkipFrame = TRUE;
		
		if (lpHw->bGotIFrame) {
		
			switch (dwTemporalReference) {                                                            
			                                                                                              
				case 0:
				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
					break;
				
				case 8:
					dwTemporalReference--;
					dwX &= ~0xFFC00000;
					dwX |= (dwTemporalReference << 22);
					lpvb->lpData[dwIndex] = ByteSwap (dwX);
					break;

				case 7:
					lpHw->bGotIFrame = FALSE;
					lpHw->bSkipFrame = FALSE;
					lpHw->Skipped++;                                                                                      
					return TRUE;
			}

		} else if (FrameType == I_FRAME) {
			lpHw->bGotIFrame = TRUE;
		}
		
		return FALSE;

	} else {
		/* for I-frame capture, just skip the current frame */
		lpHw->Skipped++;                                                                                      
		return TRUE;
	}
}


/* 
 * main interrupt service routine
 */
BOOL
HW_InterruptAcknowledge(VCUSER_HANDLE vh, BOOL FAR *lpServiced, BOOL FAR *lpRepeat)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	DWORD dwSystemTime = (pcGetTime());
	DWORD dwInterruptSpan = dwSystemTime - lpHw->dwLastInterrupt;
    DWORD Status, dwSize;

	lpHw->dwLastInterrupt = dwSystemTime;

	/* assume we should be capturing and no frame repeat needed */
    *lpServiced = TRUE;
	*lpRepeat = FALSE;

    /* verify hardware info */
    if (IsBadHWINFO(lpHw)) {
        *lpServiced = FALSE;
        return FALSE;
    }

    /* should we be captureing ??? */
    if (!lpHw->bCapture) {
        *lpServiced = FALSE;
        return FALSE;
    } 
		
    /* grab the board status */
    Status = lpHw->lpRegs->Status;

    /* deal with the event base on the overlay state */
    if (lpHw->bOverlay) {
        dwSize = HW_InterruptPassthru(vh, lpHw, Status);
    } else {
        dwSize = HW_InterruptNoPassthru(vh, lpHw, Status);
    }

    /* did we capture a good frame??? */
	if (dwSize == 0 ) {
		/* 
		 * in mpeg mode - verify we're still getting vsynch's -
		 * we may have stopped if rewinding, fast forwarding or
		 * a bad vsych
		 */
		if (lpHw->CaptureMode == CAPMODE_RTMPEG && lpHw->bVideoIn)  {
			
			if (HW_IsDead(lpHw, dwSystemTime)) {

				HW_Stop (lpHw, TRUE, FALSE);
				HW_Start (lpHw, TRUE, TRUE);
				
				RT_FlushToLastGop (lpHw);

				lpHw->bRestart = TRUE;
				lpHw->bGotIFrame = FALSE;

				lpHw->dwTimeRestart = (pcGetTime());
			} 
		}
		
        return FALSE;
    }

	if (lpHw->bRestart) {
		lpHw->bRestart = FALSE;
		lpHw->bGotIFrame = FALSE;
	}
	
    /* are we streaming??? */
    if (lpHw->bVideoIn) {
		
		DWORD dwExpectedTime, dwExpectedCount, dwCount, dwAdjustedTime;
        LPVIDEO_BUFFER lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf];
		DWORD dwTimeSpan;
		
		/* 
		 * acount for c-cube pipeline delay when captureing
		 * realtime mpeg.
		 */
		if (lpHw->CaptureMode == CAPMODE_RTMPEG && lpHw->dwInterruptCount == 0) {
			
			DWORD dwOffset = (dwSystemTime-lpHw->dwTimePrevious);

			/* 
			 * in realtime mode this is absolutely the first frame, 
			 * since we restart the microcode before capture
			 */
			if (dwOffset > lpHw->dwMicroPerInterrupt/1000) {
				lpHw->dwTimeOffset += dwOffset - (lpHw->dwMicroPerInterrupt/1000);
			} else {
				lpHw->dwTimeOffset = 0;
			}
		} 
		 
		/* 
		 * subtract any offset and calculate time since last frame 
		 */
		dwSystemTime -= lpHw->dwTimeOffset;
		dwTimeSpan = (dwSystemTime-lpHw->dwTimePrevious);

		/* 
		 * record any real missed frames - a best guess attempt at 
		 * allowing video for windows to report real dropped frames.
		 * if we don't detect it, no real problem, we'll eventually 
		 * compensate for it and remain in synch!
		 */                                                                           
		if (!lpHw->bOverlay) {
			/* 
			 * in preview mode we never miss a frame, unless we overrun the entire DMA
			 * buffer - not likely.  if that we to occur, the synch adjustment would 
			 * detect the huge expected vs. adjusted time difference and repeat frames
			 * until we were back in synch.
			 */
			dwCount = 1;
		} else {         
						  
			/* get the interrupt count */
			dwCount = TruncateingMulDiv (dwTimeSpan, 1000L, lpHw->dwMicroPerInterrupt);
			dwCount = max (dwCount, 1);

			/* 
			 * this would indicate that we toggled when we shouldn't have 
			 * (as a result of a bad vsynch -  there is a certain condition 
			 * where we saw a bad spot on a tape produce a bogus vsync
			 * which caused a bogus toggle. 
			 */
			if (dwCount == 1 && (Status & DONE_FLAG_1) && (Status & DONE_FLAG_2)) {
				
				/* 
				 * this test is only valid if didn't overrun the timer
				 * and miss clearing the bit(s)!!! (don't forget to 
				 * compensate round off errors)
				 */
				if (((dwInterruptSpan+1)*1000) < (lpHw->dwMicroPerInterrupt/2)) {
					HW_DynamicFifoReset(lpHw);
					lpHw->BadToggle++;
					return 0;
				}
			}
		}
		
		/* 
		 * record missed frames and time of the current 
		 * interrupt 
		 */
		lpHw->TimerOverRun += dwCount - 1;
		lpHw->dwTimePrevious = dwSystemTime;                                                                           

		/* adjust system time to the audio clock */
		dwAdjustedTime = ((DWORD)((double)(dwSystemTime) * lpHw->dfAudioFudgeFactor));

		/* calculate the expected interrupt count and time */                                                          
		dwExpectedCount = lpHw->dwInterruptCount + dwCount;
		dwExpectedTime = MulDiv (dwExpectedCount, lpHw->dwMicroPerInterrupt, 1000L);                                   

		/* realtime skip flag */
		if (lpHw->bSkipFrame) {
		
			/* it's time to skip a frame to correct for sync */                                                    
			if (HW_SkipFrame(lpHw, lpvb, dwSize)) {
				return FALSE;                                                                                          
			}
		/* realtime repeat flag */
		} else if (lpHw->bRepeatFrame) {
		
			/*                                                                                                     
		 	 * to double frames the time for this frame will be the 
			 * expected video time. the repeat frame function will 
			 * add 1 frame time to this (since we're one frame over 
			 * where we should be!) and keep every thing linear so 
			 * Video for Window won't report a dropped frame.                                                   
			 */                                                                                                    
			*lpRepeat = TRUE;                                                                                      

		} else if (!lpHw->bStoppingNow && !lpHw->bMultiFrame)  {
			
			/*                                                                                                             
			 * only add or drop a frame when we go out a frame so we don't                                                 
			 * go nut's adding and dropping frames because the timer ticks                                                 
			 * aren't consistant or the video rate is off a little.  We                                                    
			 * want to adjust slowly over time so it's not obvious, not                                                    
			 * constantly so the picture looks jerky and jittery.                                                          
			 */         
			DWORD dwDif = ((DWORD)abs ((int)dwAdjustedTime - (int)dwExpectedTime));
			if (dwDif >= lpHw->dwMicroPerInterrupt/1000) {                
		                                                                                                               
				/* skip or add frames to keep in sync */                                                                   
				if (dwExpectedTime > dwAdjustedTime) {                                                                     
		                                                                                                               
					/* it's time to skip a frame to correct for sync */                                                    
					if (HW_SkipFrame(lpHw, lpvb, dwSize)) {
						return FALSE;                                                                                          
					}
		                                                                                                               
				} else if (dwExpectedTime < dwAdjustedTime) {                                                              
				                                                                                                       
					/*                                                                                                     
				 	 * to double frames the time for this frame will be the 
					 * expected video time. the repeat frame function will 
					 * add 1 frame time to this (since we're one frame over 
					 * where we should be!) and keep every thing linear so 
					 * Video for Window won't report a dropped frame.                                                   
					 */                                                                                                    
					*lpRepeat = TRUE;  

					if (lpHw->CaptureMode == CAPMODE_RTMPEG) {
						/* 
						 * in real-time mode we could have been restarted due to a bad frame or
						 * fifo overflow, so we may need to insert more than one frame. By spec 
						 * the time between time stamps should be .7 seconds, and the original 
						 * gop size is 9 (. 0.300303 secs NTSC .36 secs PAL)
						 */
						lpHw->dwRepeatFrames = min ((700000/lpHw->dwMicroPerInterrupt)-9, (DWORD)MulDiv (dwDif, 1000L, lpHw->dwMicroPerInterrupt));
					} else {
						lpHw->dwRepeatFrames = 1;
					}
				}
			}
		}

    	/*                                                                                                             
		 * use the expected interrupt count to update the video time                                                   
		 * on frame time boundries.  The adjusted time is used to skip/
		 * drop frames to keep us in synch.
		 */                                                                                                            
		lpHw->dwInterruptCount = dwExpectedCount;                                                                      
		lpHw->dwVideoTime = MulDiv (lpHw->dwInterruptCount, lpHw->dwMicroPerInterrupt, 1000L);                         

	    /* 
         * we only capture if the time is right!!! 
         */
        if (lpHw->dwVideoTime >= lpHw->dwNextFrameTime || lpHw->CaptureMode == CAPMODE_RTMPEG) {

			/* advance next frame time */
            lpHw->dwNextFrameNr++;
			lpHw->dwNextFrameTime = MulDiv (lpHw->dwNextFrameNr, lpHw->dwMicroPerFrame, 1000L);

			/* mark the current buffer with the returned size */
			lpvb->dwFilled = dwSize;

			/* mark the time of capture for the cuurent frame */
			lpvb->dwTimeStamp = lpHw->dwVideoTime;

			return TRUE;
		} 
    } else if (lpHw->bCaptureFrame || lpHw->CaptureMode == CAPMODE_RTMPEG) {
        LPVIDEO_BUFFER lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf];
		BOOL bKeyFrame = FALSE;

		/* mark the current buffer with the returned size */
		lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf].dwFilled = dwSize;

        /* check for errors */
		if (Status & FIFO_OVERFLOW) {
            return FALSE;
        }

        /* 
         * transfer the frame to our the single frame 
         * buffer 
         */
        dwSize = HW_TransferFrame (lpHw
                                   ,lpHw->lpLastFrame
                                   ,lpHw->dwLastFrameMax
                                   ,lpvb
								   ,&bKeyFrame
                                   );

        /* signal it's ready */
        if (dwSize != 0) {
			
			lpHw->dwLastFrameSize = dwSize;                                                                                  

			if (lpHw->CaptureMode != CAPMODE_RTMPEG) {           
				lpHw->bCaptureFrame = FALSE;  
				SetEvent (lpHw->hFrameEvent);
			}
        }
    }

    return FALSE;
}


/*
 * Capture a frame while streaming - copy the frame buffer into 
 * the buffer passed as argument. 
 *
 * returns bytes written to lpBuffer (whose length is BufferLength)
 * called with null if no buffer - nothing to do in this case.
 *
 * returns the time stamp in lpTimeStamp (the millisec count
 * since capture began - based on the number of video syncs and the
 * video standard (PAL, NTSC...).
 */
ULONG
HW_CaptureService(
  VCUSER_HANDLE vh,
  LPUCHAR lpBuffer,
  LPULONG lpTimeStamp,
  ULONG BufferLength,
  BOOL FAR *lpKeyFrame
)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
    LPVIDEO_BUFFER lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf];
                                        
    /* verify hardware info */
    if (IsBadHWINFO(lpHw)) {
        return 0;
    }

    /* set the time stamp */
    *lpTimeStamp = lpvb->dwTimeStamp;
		
    /* tranfer the frame from hardware buffer to user buffer */
    return HW_TransferFrame (lpHw, lpBuffer, BufferLength, lpvb, lpKeyFrame);
}


/*
 * Capture a frame for single frame capture - copy the frame buffer 
 * into the buffer passed as argument. 
 *
 * returns bytes written to buffer lpBuffer (whose length is ULONG)
 * called with null if no buffer - nothing to do in this case.
 *
 */
ULONG
HW_CaptureFrame(
  VCUSER_HANDLE vh,
  LPUCHAR lpBuffer,
  ULONG BufferLength
)
{
    LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	DWORD dwSize;

	VC_TimerDisable (vh);

    /* tranfer the frame from temporary buffer to user buffer */
	dwSize = lpHw->dwLastFrameSize;

	if (BufferLength < dwSize) {
		dwSize = 0;
	} else if (dwSize != 0) {
        CopyMemory (lpBuffer, lpHw->lpLastFrame, dwSize);
    } 

	VC_TimerEnable (vh);
    
	return dwSize;
}


/*
 * Transfers a frame from lpvb input video buffer to lpOutput 
 * buffer (whose length is OutputLength). The zero padding is 
 * removed from the input buffer before the copy.
 *
 * This function takes into account that the input video buffer
 * is may be run in circular mode and rollovers can occur in the 
 * middle data.
 *  
 * returns bytes written to lpOutput buffer. Called with null if no buffer 
 * - nothing to do in this case.
 */
ULONG __inline 		 
HW_TransferFrame(
  LPHWINFO lpHw,
  LPUCHAR lpOutput,
  ULONG OutputLength,
  LPVIDEO_BUFFER lpvb,
  BOOL FAR *lpKeyFrame
)
{
    DWORD FAR *lpSearch = lpvb->lpData;
    LPUCHAR lpStart;
    
    DWORD Index = lpvb->dwStartIndex;
    DWORD dwSize = lpvb->dwFilled;
    DWORD dwStartIndex, dwChunk;
    DWORD dwIncrement = 0x10;
	DWORD dwData;
    
	*lpKeyFrame = FALSE;

    /* a size of zero indicates no data is ready */
    if (dwSize == 0) {
        return 0;
    }

    /*
     * do nothing if no output buffer is supplied
     */
    if (lpOutput == NULL) {
        return(0);
    }

    /* 
     * remove any zero padding from 
     * the start of the frame 
     */
    while (dwSize != 0) {       

		dwData = lpSearch[Index];

        /* look for the first non zero dword */
        if (dwData != 0L) {

			if (!lpHw->bVideoIn || lpHw->CaptureMode != CAPMODE_RTMPEG) {
            
	            /* if it's not a start code - bad frame!!! */
		        if (dwData != 0xb3010000) {
					
				    if (lpHw->bVideoIn) {

						/* 
						 * in overlay mode, this is an indication of possible loss of
						 * vsynch due to a "bad" spot in the source video - try to
						 * recover...
						 */
						if (lpHw->bOverlay) 
							HW_DynamicFifoReset(lpHw);

					    lpHw->NoStartCode++;
					}

					return 0;
				} 

	            /* now we know the start of the frame... */
		        lpStart = (LPUCHAR)(lpSearch + Index);
			    dwStartIndex = Index;
				*lpKeyFrame = TRUE;
				break;

            } else {

		        if (dwData == 0xb3010000) {
					*lpKeyFrame = TRUE;
				}

				dwStartIndex  = lpvb->dwStartIndex;
				dwSize = lpvb->dwFilled;
				lpStart = (LPUCHAR)(lpvb->lpData+dwStartIndex);
				break;
			}
        }
        
        dwSize -= min (dwIncrement, dwSize);
	    Index = (Index + min (dwIncrement, dwSize)) % lpvb->dwLength;
    }

	/* frame should always be on 16 dword boundries */
	if (dwSize % 0x10 != 0) {

		if (lpHw->bVideoIn) {
			
			/* 
			 * should never happen in preview mode, but in overlay mode 
			 * this is an indication of possible loss of vsynch due to 
			 * a "bad" spot in the source video - try to recover...
			 */
			if (lpHw->bOverlay) {
				HW_DynamicFifoReset(lpHw);
			}
			
			if (lpHw->CaptureMode == CAPMODE_RTMPEG) {
				lpHw->bRestart = TRUE;
				lpHw->bGotIFrame = TRUE;
				RT_FlushToLastGop (lpHw);
			}

			lpHw->ExtractFrame++;
			return 0;
        }
	}

	/* clear lost frame count */
	lpHw->dwLostFrames = 0L;

    /* 
     * if we got this far and size is 0, then buffer is full of all zero's.
     * this is a bogus situation, and we should report it as 
     * a bad frame
     */
    if (dwSize == 0) {
        if (lpHw->bVideoIn) {
			/* 
	  	     * in overlay mode, this is an indication of possible loss of
			 * vsynch due to a "bad" spot in the source video - try to
			 * recover...
			 */
			if (lpHw->bOverlay) 
				HW_DynamicFifoReset(lpHw);

            lpHw->FrameAllZeros++;

			if (lpHw->CaptureMode == CAPMODE_RTMPEG) {
				lpHw->bRestart = TRUE;
				lpHw->bGotIFrame = TRUE;
				RT_FlushToLastGop (lpHw);
			}

        }
        return 0;
    } 

    /* now convert from dwords to bytes */
	dwSize *= sizeof(DWORD);

    /* fail if the output buffer is too small */
    if (dwSize > OutputLength || dwSize > (lpHw->dwMaxCalcFrameSize*sizeof(DWORD))) {
        if (lpHw->bVideoIn) {
            lpHw->FrameTooBig++;
        }
        return 0;
    }

	/* copy frame to output buffer */
	dwChunk = min (dwSize, (lpvb->dwLength - dwStartIndex) * sizeof(DWORD));
	CopyMemory (lpOutput, lpStart, dwChunk);

	if (dwSize != dwChunk) {
		CopyMemory (lpOutput+dwChunk, lpSearch, dwSize-dwChunk);
	}

	/* check for buffer overrun */
    if (lpHw->bOverlay) {
	    
		DWORD Status = lpHw->lpRegs->Status;
		BOOL bFailed = FALSE;

		if (Status & DONE_FLAG_2 && lpHw->CaptBufs.dwNumberBufs == 1) {
			bFailed = TRUE;				
		} else {
			DWORD dwStart = lpHw->lpRegs->MasterStart2;
			DWORD dwPhys = lpvb->dwPhys;
	
			if (dwStart > dwPhys && dwStart < dwPhys + MAX_CAPT_SIZE*sizeof(DWORD))	{
				lpHw->lpRegs->Status = DONE_FLAG_1+DONE_FLAG_2;
				bFailed = TRUE;
			}
		}

		/* check to see if we have overwritten data in the buffer */
		if (bFailed) {        
		    if (lpHw->bVideoIn) {
		        lpHw->CopyOverRun++;
			}
			return 0;
		}

	} else if (HW_CheckPreviewOverRun(lpHw, lpvb)) {
		return 0;
	}

	/* 
     * store the actual frame size in place of the 
     * start code for the codec
     */
	if (!lpHw->bVideoIn || lpHw->CaptureMode != CAPMODE_RTMPEG) {
	    *((DWORD FAR*)lpOutput) = 0xff000000|dwSize;
	}

	/* return the bytes copied */
	return dwSize;
}



