/*
 * vcuser.c
 *
 * 32-bit Video Capture driver
 * User-mode support library
 *
 * functions providing access to video capture hardware, 
 * by accessing the kernel-mode device driver. 
 * 
 *
 * Mark Vrabel, October 95:
 */

#define _GLOBAL_INIT_


#include <vcdriver.h>
#include <vtdioc.h>
#include "vcupriv.h"
#include "registry.h"


/*
 * hardware open count for this process
 */ 
static DWORD gdwOpenCount = 0;


/*
 * DLLMain
 *
 * Called when DLL is loaded by a process, and when new threads are 
 * created by a process that has already loaded the DLL.  Also called 
 * when threads of a process that has loaded the DLL exit cleanly and 
 * when the process itself unloads the DLL.
 */

BOOL WINAPI DllMain(HINSTANCE hDLLInst, DWORD fdwReason, LPVOID lpvReserved)
{   
    switch (fdwReason) {
        
        case DLL_PROCESS_ATTACH:
            ghDLLInst = hDLLInst;
            break;

        case DLL_PROCESS_DETACH:
             break;

        case DLL_THREAD_ATTACH:
            break;

        case DLL_THREAD_DETACH:
            break;
    }
    return TRUE;
}


/*
 * open the device and return a capture device handle that can be used
 * in future calls.
 * 
 * For Win95:
 *  The device index indicates a specific plug and play device node 
 *  to open supported by the lpDriverName kernel mode driver. 
 *
 * For WinNT:
 *  The device index is 0 for the first capture device up to N for the
 *  Nth installed capture device.
 *
 *  If lpDriverName is non-null, then we will open the Nth device handled
 *  by this driver. Current implementation supports only one device per
 *  drivername. 
 *
 * This function returns NULL if it is not able to open the device.
 */                       
VCUSER_HANDLE WINAPI
VC_OpenDevice(LPTSTR lpDriverName, DWORD DeviceIndex)
{
    VCUSER_HANDLE vh;

	/*
	 * we allow multiple board support, but only one per
	 * process!!!
	 */
	if (gdwOpenCount != 0) {
		return NULL;
	}

    /* try to initialize/detect the hardware */
    vh = DeviceEntry(lpDriverName, DeviceIndex);
    if (IsBadVCUSER_HANDLE(vh)) {
        return NULL;
    }

    /* call the hardware specific open function */
    if (vh->Callback.DeviceOpenFunc != NULL) {
        if (!vh->Callback.DeviceOpenFunc(vh)) {
            VC_CloseDevice(vh);
            return NULL;
        }
    }

	/* increment the procecss open count */
	gdwOpenCount++;
    return(vh);
}


VOID WINAPI
VC_CloseDevice(VCUSER_HANDLE vh) 
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return;
    }

    /* call the hardware specific close function */
    if (vh->Callback.DeviceCloseFunc != NULL) {
        vh->Callback.DeviceCloseFunc(vh);
    }

    /* clean up everything else */
    VC_Cleanup(vh);

	/* decrement the procecss open count */
//		gdwOpenCount = (gdwOpenCount == 0) ?0:gdwOpenCount--;

	if(gdwOpenCount>0)
		gdwOpenCount--;
}


/*
 * Configuration.
 *
 * These functions perform device-dependent setup affecting the
 * target format, the source acquisition or the display (overlay).
 *
 * The structures passed are not interpreted by the vcuser and vckernel
 * libraries except that the first ulong of the struct must contain the
 * size in bytes of the entire structure (see vcstruct.h). It is assumed
 * that the structures are defined and agreed between the user-mode
 * hardware-specific code and the kernel-mode hardware specific code
 */
BOOL WINAPI
VC_ConfigFormat(VCUSER_HANDLE vh, LPCONFIG_INFO lpGeneric)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.ConfigFormatFunc != NULL) {
        return (vh->Callback.ConfigFormatFunc(vh, lpGeneric));
    } 

    return FALSE;
}


BOOL WINAPI
VC_ConfigSource(VCUSER_HANDLE vh, LPCONFIG_INFO lpGeneric)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.ConfigSourceFunc != NULL) {
        return (vh->Callback.ConfigSourceFunc(vh, lpGeneric));
    } 

    return FALSE;
}


BOOL WINAPI
VC_ConfigDisplay(VCUSER_HANDLE vh, LPCONFIG_INFO lpGeneric)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.ConfigDisplayFunc != NULL) {
        return (vh->Callback.ConfigDisplayFunc(vh, lpGeneric));
    } 

    return FALSE;
}


/*
 * overlay and keying
 *
 * Several different methods are used by devices to locate the overlay
 * area on the screen: colour (either rgb or palette index) and/or
 * either a single rectangle, or a series of rectangles defining a complex
 * region. Call GetOverlayMode first to find out which type of overlay
 * keying is available. If this returns 0, this hardware is not capable
 * of overlay.
 */

/*
 * find out the overlay keying method
 */
ULONG WINAPI
VC_GetOverlayMode(VCUSER_HANDLE vh)
{
    ULONG ulMode = 0;

    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return 0;
    }

    if (vh->Callback.GetOverlayModeFunc != NULL) {
        ulMode = vh->Callback.GetOverlayModeFunc(vh);
    } 

    return ulMode;
}


/*
 * set the key colour to a specified RGB colour. This function will only
 * succeed if GetOverlayMode returned VCO_KEYCOLOUR and VCO_KEYCOLOUR_RGB
 * and not VCO_KEYCOLOUR_FIXED
 */
BOOL WINAPI
VC_SetKeyColourRGB(VCUSER_HANDLE vh, LPRGBQUAD lpRGB)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.SetKeyRGBFunc != NULL) {
        return (vh->Callback.SetKeyRGBFunc(vh, lpRGB));
    } 

    return FALSE;
}


/*
 * set the key colour to a specified palette index. This function will only
 * succeed if GetOverlayMode returned VCO_KEYCOLOUR and not either
 * VCO_KEYCOLOUR_RGB or VCO_KEYCOLOUR_FIXED
 */
BOOL WINAPI
VC_SetKeyColourPalIdx(VCUSER_HANDLE vh, WORD PalIndex)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.SetKeyPalIdxFunc != NULL) {
        return (vh->Callback.SetKeyPalIdxFunc(vh, PalIndex));
    } 

    return FALSE;
}


/*
 * get the current key colour. This 32-bit value should be interpreted
 * as either a palette index or an RGB value according to the
 * VCO_KEYCOLOUR_RGB flag returned from VC_GetOverlayMode.
 */
DWORD WINAPI
VC_GetKeyColour(VCUSER_HANDLE vh)
{
    ULONG ulKey = 0;

    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return 0;
    }

    if (vh->Callback.GetKeyColourFunc != NULL) {
        ulKey = vh->Callback.GetKeyColourFunc(vh);
    }

    return ulKey;
}


/*
 * Set the overlay parent window
 */
BOOL WINAPI
VC_SetOverlayWindow(VCUSER_HANDLE vh, HWND hWnd)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.SetOverlayWindowFunc != NULL) {
        return (vh->Callback.SetOverlayWindowFunc(vh, hWnd));
    } 

    return FALSE;
}


/*
 * set the overlay rectangle(s). This rectangle marks the area in device
 * co-ordinates where the overlay video will appear. The video will be
 * panned so that pixel (0,0) will appear at the top-left of this rectangle,
 * and the video will be cropped at the bottom and right.  The video
 * stream will not normally be scaled to fit this window: scaling is normally
 * determined by the destination format set by VC_ConfigFormat.
 *
 * If VCO_KEYCOLOUR was returned, the video
 * will only be shown at those pixels within the rectangle for which the
 * vga display has the key colour (VC_GetKeyColour() for this).
 *
 * Some devices may support complex regions (VCO_COMPLEX_RECT). In that case,
 * the first rectangle in the area must be the bounding rectangle for
 * the overlay area, followed by one rectangle for each region within it in
 * which the overlay should appear.
 */
BOOL WINAPI
VC_SetOverlayRect(VCUSER_HANDLE vh, LPOVERLAY_RECTS lpOR)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /*
     * the structure OVERLAY_RECTS can include one or more rectangles.
     * There must be at least one, and the structure as defined includes
     * one rectangle
     */
    if (lpOR->ulCount < 1) {
        return(FALSE);
    }

    if (vh->Callback.SetOverlayRectsFunc != NULL) {
        return (vh->Callback.SetOverlayRectsFunc(vh, lpOR));
    } 

    return FALSE;
}


/*
 * set the offset of the overlay. This changes the panning - ie which
 * source co-ordinate appears as the top left pixel in the overlay rectangle.
 * Initially after a call to VC_SetOverlayRect, the source image will be panned
 * so that the top-left of the source image is aligned with the top-left of the
 * overlay rectangle. This call aligns the top-left of the source image
 * with the top-left of this offset rectangle.
 */
BOOL WINAPI
VC_SetOverlayOffset(VCUSER_HANDLE vh, LPRECT lprc)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.SetOverlayOffsetFunc != NULL) {
        return (vh->Callback.SetOverlayOffsetFunc(vh, lprc));
    } 

    return FALSE;
}


/* enable or disable overlay. if the BOOL bOverlay is TRUE, and the overlay
 * key colour and rectangle have been set, overlay will be enabled.
 */
BOOL WINAPI
VC_Overlay(VCUSER_HANDLE vh, BOOL bOverlay)
{
    BOOL bOK;

    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.OverlayFunc != NULL) {
        bOK = (vh->Callback.OverlayFunc(vh, bOverlay));
    } 

    return FALSE;
}


/*
 * enable or disable acquisition.
 * If acquisition is disabled, the overlay image will be frozen.
 *
 * this function will have no effect during capture since the acquisition
 * flag is toggled at each frame capture.
 */
BOOL WINAPI
VC_Capture(VCUSER_HANDLE vh, BOOL bAcquire)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.CaptureFunc != NULL) {
        return (vh->Callback.CaptureFunc(vh, bAcquire));
    } 

    return FALSE;
}


/*
 * write a frame to the frame buffer for playback
 */
BOOL WINAPI
VC_DrawFrame(VCUSER_HANDLE vh, LPDRAWBUFFER lpDraw)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    if (vh->Callback.DrawFrameFunc != NULL) {
        return (vh->Callback.DrawFrameFunc(vh, lpDraw));
    } 

    return FALSE;
}


/*
 * capture a single frame, synchronously. the video header must point
 * to a data buffer large enough to hold one frame of the format set by
 * VC_ConfigFormat.
 *
 */
BOOL WINAPI
VC_Frame(VCUSER_HANDLE vh, LPVIDEOHDR lpVideoHdr)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* check we're not busy  */
    if (vh->State == State_Start) {
        return FALSE;
    }

    /* verify we have a buffer */
    if (lpVideoHdr == NULL) {
        return FALSE;
    }

    /* verify buffer size */
    if (lpVideoHdr->dwBufferLength < vh->ImageSize) {
        return(FALSE);
    }

    if (vh->Callback.FrameFunc != NULL) {
        return (vh->Callback.FrameFunc(vh, lpVideoHdr));
    } 

    return FALSE;
}


/*
 * prepare to start capturing frames -
 * create the worker thread and wait for it to signal completion of its
 * setup.
 */
BOOL  WINAPI
VC_StreamInit(
    VCUSER_HANDLE vh,
    LPVCCALLBACK lpCallback,  // pointer to callback function
    ULONG ulFrameRate         // desired capture rate: microseconds per frame
)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* check we're not busy  */
    if (vh->State != State_Idle) {
        return FALSE;
    }
   
    /* store callback for use by worker thread */
    CopyMemory (&vh->UserCallback, lpCallback, sizeof(VCCALLBACK));
       
    /* frame rate is the argument for the init function */
    vh->FunctionArg  = (DWORD) ulFrameRate;

    /* create the thread */
    vh->hThread = CreateThread(
            NULL,
            0,
            (LPTHREAD_START_ROUTINE) VC_ThreadInit,
            vh,
            0,
            &vh->ThreadId);

    if (vh->hThread == NULL) {
        return(FALSE);
    }

    /* wait for completion of the init operation */
    WaitForSingleObject(vh->hCompleteEvent, INFINITE);

    /* check that the init succeeded */
    if ( (BOOL) vh->FunctionResult == TRUE) {
        BOOL ret = SetThreadPriority(vh->hThread, THREAD_PRIORITY_HIGHEST);
        return(TRUE);
    }

    /* init failed - wait for the thread to exit and clean up */
    WaitForSingleObject(vh->hThread, INFINITE);

    CloseHandle(vh->hThread);
    vh->hThread = NULL;

    return(FALSE);
}
                       

/*
 * clean up after capturing. You must have stopped capturing first.
 */
BOOL WINAPI
VC_StreamFini(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* make sure we're not busy */
    /* check we're not busy - a STOP should have occured */
    if (vh->State != State_Init) {
        return FALSE;
    }                          

    /* can't call stream-fini from a callback. not nice */
    if (GetCurrentThreadId() == vh->ThreadId) {
        return(FALSE);
    }

    /* or have any buffers on the queue */
    if (vh->ReadyListHead != NULL) {
        return(FALSE);
    }
    if (vh->DoneListHead != NULL) {
        return(FALSE);
    }

    if (vh->Callback.StreamFiniFunc != NULL) {
        if (vh->Callback.StreamFiniFunc(vh)) {

            /* set the state and alert the thread */
            vh->State = State_Idle;

	        SetThreadPriority(vh->hThread, THREAD_PRIORITY_HIGHEST);
            SetEvent(vh->hWorkEvent);

            /* on this function, we don't wait for the completion event -
             * the thread just cleans up and exits. We wait for the thread to
             * terminate
             */
            WaitForSingleObject(vh->hThread, INFINITE);

            CloseHandle(vh->hThread);
            vh->hThread = NULL;

            return TRUE;
        }
    }

    return FALSE;
}


/*
 * initiate capturing of frames. Must have called VC_StreamInit first.
 */
BOOL WINAPI
VC_StreamStart(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* check that an INIT has already been done */
    if (vh->State != State_Init) {
        return FALSE;
    }
    
    if (vh->Callback.StreamStartFunc != NULL) {

		vh->SkipCount = 0;
        if (vh->Callback.StreamStartFunc(vh)) {
            vh->State = State_Start;
            return TRUE;
        } 
    }

    return FALSE;
}


/*
 * stop capturing frames. Current frame may still complete. All other buffers
 * will remain in the queue until capture is re-started, or they are released
 * by VC_StreamReset.
 *
 * again we do not need to go via the worker thread for this.
 */
BOOL WINAPI
VC_StreamStop(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* check that we are not idle (Init must have been done,
     * though not necessarily Start)
     */
    if (vh->State == State_Idle) {
        return FALSE;
    }

    if (vh->Callback.StreamStopFunc != NULL) {

        if (vh->Callback.StreamStopFunc(vh)) {
            vh->State = State_Init;
            return TRUE;
        } 
    }

    return FALSE;
}


/*
 * cancel all buffers that have been 'add-buffered' but have not
 * completed. This will also force VC_StreamStop if it hasn't already been
 * called.
 *
 * This request needs to go to the worker thread, since all but a few (2)
 * of the buffers will be queued up in user-space by the worker thread.
 */
BOOL WINAPI
VC_StreamReset(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* must be Init-ed or Start-ed, not idle */
    if (vh->State == State_Idle) {
        return FALSE;
    }

    /* stop the streaming */
    if (vh->State == State_Start) {
        VC_StreamStop(vh);
    }

    /*
     * complete all the add-buffer and done requests on the queues
     */
    VC_ResetDoneQueue(vh);
    VC_ResetReadyQueue(vh);
    
    vh->ReadyListHead = NULL;
    vh->ReadyListTail = NULL;
    vh->DoneListHead  = NULL;
    vh->DoneListTail  = NULL;

    return TRUE;
}


/*
 * get the count of frames that have been skipped since the last call
 * to VC_StreamInit.
 *
 * The worker thread keeps count of the number of skips. It is easier
 * to synchronize access to the count by treating this as a function,
 * though it might be more efficient just to grab the count
 * ourselves from shared data.
 */
ULONG WINAPI
VC_GetStreamError(VCUSER_HANDLE vh)
{
    ULONG ulCount;

    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return 0;
    }

    /* check that we are not idle (Init must have been done,
     * though not necessarily Start)
     */
    if (vh->State == State_Idle) {
        return FALSE;
    }

    ulCount = vh->SkipCount;
    vh->SkipCount = 0;
    
    /* return the result */
    return(ulCount);
}


/*
 * add a buffer to the queue. The buffer should be large enough
 * to hold one frame of the format specified by VC_ConfigFormat.
 *
 * this function goes to the worker thread, who will queue buffers in user
 * space and pass a limited number on to the kernel driver
 */
BOOL WINAPI
VC_StreamAddBuffer(VCUSER_HANDLE vh, LPVIDEOHDR lpVideoHdr)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* check that we are not idle (Init must have been done,
     * though not necessarily Start)
     */
    if (vh->State == State_Idle) {
        return FALSE;
    }

    /* verify we have a buffer */
    if (lpVideoHdr == NULL) {
        return FALSE;
    }

    /* verify buffer size */
    if (lpVideoHdr->dwBufferLength < vh->ImageSize) {
        return(FALSE);
    }

    /* Return error if buffer has not been prepared. */
    if (!(lpVideoHdr->dwFlags & VHDR_PREPARED)) {
        return(FALSE);
    }

    /* Return error if buffer is already in the queue. */
    if (lpVideoHdr->dwFlags & VHDR_INQUEUE) {
        return(FALSE);
    }

    /* OK to add it to the queue */
    VC_AddReadyTail(vh, lpVideoHdr);

    /* return the result */
    return TRUE;
}


/*
 * get the current position within the capture stream (ie time
 * in millisecs since capture began).
 *
 * This has no interaction with the worker thread and can go straight to
 * the device.
 */
BOOL WINAPI
VC_GetStreamPos(VCUSER_HANDLE vh, LPMMTIME pTime)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return FALSE;
    }

    /* check that we are not idle (Init must have been done,
     * though not necessarily Start)
     */
    if (vh->State == State_Idle) {
        return FALSE;
    }

    if (vh->Callback.StreamGetPositionFunc != NULL) {
        ULONG ulPosition;
    
        ulPosition = vh->Callback.StreamGetPositionFunc(vh);
        pTime->wType = TIME_MS;
        pTime->u.ms = (DWORD) ulPosition/1000L;

        return TRUE;
    } 

    return FALSE;
}
